package jlx.behave.proto;

import java.time.LocalTime;
import java.util.*;

import jlx.asal.j.*;
import jlx.asal.rewrite.ASALSymbolicValue;
import jlx.asal.vars.ASALProperty;
import jlx.asal.vars.ASALVariable;
import jlx.behave.stable.DecaStableOutputEvolution;
import jlx.models.UnifyingBlock.ReprPort;
import jlx.utils.*;
import jlx.utils.ldd.LDD;
import jlx.utils.ldd.TransitiveLDD;

public class DecaFourStateConfig implements LineageObject {
	private final Map<JScope, DecaFourVertex> vtxs;
	private final int vtxsHashCode;
	
	public int level;
	
	public DecaFourStateConfig(Map<JScope, DecaFourVertex> vtxs) {
		this.vtxs = vtxs;
		
		vtxsHashCode = Objects.hash(vtxs);
	}
	
	public Map<String, String> getValuePerOutput() {
		Map<String, String> result = new TreeMap<String, String>();
		
		for (DecaFourVertex v : vtxs.values()) {
			for (Map.Entry<ReprPort, ASALSymbolicValue> e : v.getOutputVal().extractValuation().entrySet()) {
				String key = e.getKey().getReprOwner().getName() + "::" + e.getKey().getName();
				String value = e.getValue().toString();
				result.put(key, value);
			}
		}
		
		return result;
	}
	
	public Set<Class<?>> getSysmlClzs() {
		Set<Class<?>> result = new HashSet<Class<?>>();
		
		for (DecaFourVertex v : vtxs.values()) {
			result.addAll(v.getSysmlClzs());
		}
		
		return result;
	}
	
	public Map<ASALProperty, ASALSymbolicValue> getPropertyVal() {
		Map<ASALProperty, ASALSymbolicValue> result = new HashMap<ASALProperty, ASALSymbolicValue>();
		
		for (DecaFourVertex v : vtxs.values()) {
			for (Map.Entry<ASALVariable, ASALSymbolicValue> e : v.getLegacy().getLegacy().getStateConfig().getValuation().entrySet()) {
				if (e.getKey() instanceof ASALProperty) {
					result.put((ASALProperty)e.getKey(), e.getValue());
				}
			}
		}
		
		return result;
	}
	
	public PulsePackMap getOutputVal() {
		return extractOutputVal();
	}
	
	private PulsePackMap extractOutputVal() {
		Map<ReprPort, PulsePack> result = new HashMap<ReprPort, PulsePack>();
		
		for (DecaFourVertex v : vtxs.values()) {
			result.putAll(v.getOutputVal().getPackPerPort());
		}
		
		return new PulsePackMap(result, Dir.OUT);
	}
	
	public Map<JScope, DecaFourVertex> getVtxs() {
		return vtxs;
	}
	
	/**
	 * Gives all successors that can be reached in one step via the given (external!) inputs.
	 * Pulses in the given inputs are ignored (in other words, given inputs are deactivated).
	 * @param deactivatedExternalInputs Inputs.
	 * @param sicInputs Exists for debug purposes only.
	 */
	public Map<DecaFourStateConfig, DecaFourCoverageStats> getSuccsViaInputs(PulsePackMap deactivatedExternalInputs, PulsePackMap sicInputs) {
		Map<DecaFourVertex, Set<DecaFourTransition>> trsPerTgt = new HashMap<DecaFourVertex, Set<DecaFourTransition>>();
		Map<JScope, Set<DecaFourVertex>> pssPerScope = new HashMap<JScope, Set<DecaFourVertex>>(); //Possible SuccessorS
		int index = 0;
		
		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
			Set<DecaFourVertex> pss = new HashSet<DecaFourVertex>();
			
			for (Map.Entry<DecaFourTransitionReqSet, Map<PulsePackMap, Set<DecaFourTransition>>> e2 : e1.getValue().getTrsPerNonDetTgtsPerReqSet().entrySet()) {
				if (e2.getKey().containsMatch(getVtxs())) {
//					System.out.println("option " + index);
					index++;
					
					for (Map.Entry<PulsePackMap, Set<DecaFourTransition>> e3 : e2.getValue().entrySet()) {
						if (deactivatedExternalInputs.implies(e3.getKey())) {
							for (DecaFourTransition t : e3.getValue()) {
								for (DecaFourVertex tgt : t.getTgtGrp().getVtxs()) {
									HashMaps.inject(trsPerTgt, tgt, t);
								}
								
								pss.addAll(t.getTgtGrp().getVtxs());
							}
							
//							protoTrsDest.addAll(e1.getValue().getProtoTrsPerReqSet().get(
//							System.out.println("adding " + e3.getKey().getVtxs().size());
						}
					}
				}
			}
			
			if (pss.isEmpty()) {
				throw new Error("Should not happen!");
			}
			
			pssPerScope.put(e1.getKey(), pss);
		}
		
//		System.out.println("#combis = " + HashMaps.getCombinationCount(pssPerScope, false));
		
		Map<DecaFourStateConfig, DecaFourCoverageStats> result = new HashMap<DecaFourStateConfig, DecaFourCoverageStats>();
		
		for (Map<JScope, DecaFourVertex> perm : HashMaps.allCombinations(pssPerScope)) {
			DecaFourCoverageStats covStats = new DecaFourCoverageStats();
			
			for (DecaFourVertex tgt : perm.values()) {
				covStats.add4(trsPerTgt.get(tgt));
			}
			
			result.put(new DecaFourStateConfig(perm), covStats);
		}
		
		return result;
	}
	
	/**
	 * Gives a sequence of successors that:
	 *  (1) is the only one that can be initiated with the given inputs;
	 *  (2) ends with a stable successor;
	 *  (3) does not pass through cycles.
	 * If NULL, no such sequence exists.
	 * Pulses in the given inputs are ignored (in other words, given inputs are deactivated).
	 * @param deactivatedExternalInputs Inputs.
	 * @param sicInputs Exists for debug purposes only.
	 */
	public static DecaFourStateConfigPath followInputs(Set<DecaFourStateConfig> start, PulsePackMap deactivatedExternalInputs, PulsePackMap sicInputs) {
		//DecaFourStateConfig endPoint = null;
		DecaFourStateConfigPath basePathToEndPoint = null;
		Set<DecaStableOutputEvolution> baseEvos = null;
		Map<JScope, List<DecaFourVertex>> basePathPerScopeToEndPoint = null;
		
		Set<DecaFourStateConfigPath> paths = new HashSet<DecaFourStateConfigPath>();
		Set<DecaFourStateConfigPath> newPaths = new HashSet<DecaFourStateConfigPath>();
		
		for (DecaFourStateConfig cfg : start) {
			paths.add(new DecaFourStateConfigPath(cfg, PulsePackMap.EMPTY_IN, 0));
		}
		
		if ("disable".equals("off")) {
			throw new Error("DISABLED");
		}
		
//		System.out.println("[" + LocalTime.now() + "] #paths = " + paths.size());
		
		while (paths.size() > 0) {
			newPaths.clear();
			
			for (DecaFourStateConfigPath path : paths) {
				Map<DecaFourStateConfig, DecaFourCoverageStats> succsViaInputs = path.getLastCfg().getSuccsViaInputs(deactivatedExternalInputs, sicInputs);
//				System.out.println("#succs = " + succsViaInputs.size() + "; len = " + path.size());
				
				if (succsViaInputs.size() == 1) {
					DecaFourStateConfig succ = succsViaInputs.keySet().iterator().next();
					
					if (succ.equals(path.getLastCfg())) {
						if (basePathToEndPoint != null) {
							if (path.getLastCfg().equals(basePathToEndPoint.getLastCfg())) {
								if (isCompatiblePath(basePathPerScopeToEndPoint, baseEvos, path.getCfgs())) {
									//Do nothing.
								} else {
									if (sicInputs != null) {
										System.out.println("Incompatible path!");
										System.exit(0);
									}
									
									return null;
								}
							} else {
								if (sicInputs != null) {
									System.out.println("Different end point!");
									System.out.println("Path 1:");
									
									for (int i = 0; i < path.getCfgs().size(); i++) {
										System.out.println(" [" + i + "] " + path.getCfgs().get(i).getDescription());
									}
									
									System.out.println("Path 2:");
									
									for (int i = 0; i < basePathToEndPoint.getCfgs().size(); i++) {
										System.out.println(" [" + i + "] " + basePathToEndPoint.getCfgs().get(i).getDescription());
									}
									
									System.out.println("Input:");
									System.out.println(" " + deactivatedExternalInputs.toString().replace(" | ", "\n "));
									
									System.exit(0);
								}
								
								return null;
							}
						} else {
//							endPoint = lastCfg;
							basePathToEndPoint = path;
//							basePathPerScopeToEndPoint = extractPathPerScope(path);
							baseEvos = DecaStableOutputEvolution.getOutputEvolutions(path.getCfgs());
						}
					} else {
						if (path.getCfgs().contains(succ)) {
							if (sicInputs != null) {
								System.out.println("Loop!");
								System.exit(0);
							}
							
							return null;
						} else {
							newPaths.add(new DecaFourStateConfigPath(path, succ, succsViaInputs.values().iterator().next(), PulsePackMap.EMPTY_IN, 0));
						}
					}
				} else { // [ succsViaInputs.size() > 1 ]
					for (Map.Entry<DecaFourStateConfig, DecaFourCoverageStats> e : succsViaInputs.entrySet()) {
						if (path.getCfgs().contains(e.getKey())) {
							System.out.println(deactivatedExternalInputs.toString());
							
							for (DecaFourStateConfig x : path.getCfgs()) {
								System.out.println(x.getDescription());
							}
							
							for (DecaFourStateConfig x : succsViaInputs.keySet()) {
								System.out.println("Available: " + x.getDescription());
							}
							throw new Error("Infinite loop!");
						}
						
						newPaths.add(new DecaFourStateConfigPath(path, e.getKey(), e.getValue(), PulsePackMap.EMPTY_IN, 0));
					}
				}
			}
			
			paths.clear();
			paths.addAll(newPaths);
			
//			System.out.println("[" + LocalTime.now() + "] #paths = " + paths.size());
		}
		
		if (basePathToEndPoint == null) {
			throw new Error("Should not happen!");
		}
		
		return basePathToEndPoint;
	}
	
	private boolean extendPathViaInputs(DecaFourStateConfigPath path, PulsePackMap input, int nextInputValId, Set<DecaFourStateConfigPath> finishedDest, Set<DecaFourStateConfigPath> newPaths) {
		Map<DecaFourStateConfig, DecaFourCoverageStats> succsViaInputs = path.getLastCfg().getSuccsViaInputs(input, null);
		
		for (Map.Entry<DecaFourStateConfig, DecaFourCoverageStats> e : succsViaInputs.entrySet()) {
			if (path.getLastCfg().equals(e.getKey())) {
				finishedDest.add(path); //Stabilized!
			} else {
				if (path.getCfgs().contains(e.getKey())) {
					return true; //Divergence!
				} else {
					newPaths.add(new DecaFourStateConfigPath(path, e.getKey(), e.getValue(), PulsePackMap.EMPTY_IN, 0));
				}
			}
		}
		
		return false;
	}
	
	public static Set<DecaFourStateConfigPath> computeStabilizingPaths(Set<DecaFourStateConfig> start, PulsePackMap deactivatedExternalInputs, PulsePackMap sicInputs) {
		Set<DecaFourStateConfigPath> result = new HashSet<DecaFourStateConfigPath>();
		Set<DecaFourStateConfigPath> paths = new HashSet<DecaFourStateConfigPath>();
		Set<DecaFourStateConfigPath> newPaths = new HashSet<DecaFourStateConfigPath>();
		
		for (DecaFourStateConfig cfg : start) {
			paths.add(new DecaFourStateConfigPath(cfg, PulsePackMap.EMPTY_IN, 0));
		}
		
		while (paths.size() > 0) {
			newPaths.clear();
			
			for (DecaFourStateConfigPath path : paths) {
				Map<DecaFourStateConfig, DecaFourCoverageStats> succsViaInputs = path.getLastCfg().getSuccsViaInputs(deactivatedExternalInputs, sicInputs);
				
				for (Map.Entry<DecaFourStateConfig, DecaFourCoverageStats> e : succsViaInputs.entrySet()) {
					if (path.getLastCfg().equals(e.getKey())) {
						result.add(path); //Stabilized!
					} else {
						if (path.getCfgs().contains(e.getKey())) {
//							System.out.println("divergence!");
							return Collections.emptySet(); //Divergence!
						} else {
							newPaths.add(new DecaFourStateConfigPath(path, e.getKey(), e.getValue(), PulsePackMap.EMPTY_IN, 0));
						}
					}
				}
			}
			
			paths.clear();
			paths.addAll(newPaths);
		}
		
		return result;
	}
	
	public static Set<DecaFourStateConfigPath> getDivergingPaths(Set<DecaFourStateConfig> start, PulsePackMap deactivatedExternalInputs, PulsePackMap sicInputs) {
		Set<DecaFourStateConfigPath> result = new HashSet<DecaFourStateConfigPath>();
		Set<DecaFourStateConfigPath> paths = new HashSet<DecaFourStateConfigPath>();
		Set<DecaFourStateConfigPath> newPaths = new HashSet<DecaFourStateConfigPath>();
		
		for (DecaFourStateConfig cfg : start) {
			paths.add(new DecaFourStateConfigPath(cfg, PulsePackMap.EMPTY_IN, 0));
		}
		
		while (paths.size() > 0) {
			newPaths.clear();
			
			for (DecaFourStateConfigPath path : paths) {
				Map<DecaFourStateConfig, DecaFourCoverageStats> succsViaInputs = path.getLastCfg().getSuccsViaInputs(deactivatedExternalInputs, sicInputs);
				
				for (Map.Entry<DecaFourStateConfig, DecaFourCoverageStats> e : succsViaInputs.entrySet()) {
					if (path.getLastCfg().equals(e.getKey())) {
						//Stabilized!
					} else {
						if (path.getCfgs().contains(e.getKey())) {
							result.add(new DecaFourStateConfigPath(path, e.getKey(), e.getValue(), PulsePackMap.EMPTY_IN, 0));
						} else {
							newPaths.add(new DecaFourStateConfigPath(path, e.getKey(), e.getValue(), PulsePackMap.EMPTY_IN, 0));
						}
					}
				}
			}
			
			paths.clear();
			paths.addAll(newPaths);
		}
		
		return result;
	}
	
	public static Set<DecaFourStateConfigPath> followInputs2(Set<DecaFourStateConfig> start, PulsePackMap deactivatedExternalInputs, PulsePackMap sicInputs) {
		Map<DecaFourStateConfig, Set<DecaFourStateConfigPath>> pathsPerTgt = new HashMap<DecaFourStateConfig, Set<DecaFourStateConfigPath>>();
		Set<DecaFourStateConfigPath> paths = computeStabilizingPaths(start, deactivatedExternalInputs, sicInputs);
//		System.out.println("#paths = " + paths.size());
		
		for (DecaFourStateConfigPath path : paths) {
			HashMaps.inject(pathsPerTgt, path.getLastCfg(), path);
		}
		
		Set<DecaFourStateConfigPath> result = new HashSet<DecaFourStateConfigPath>();
		
		for (Map.Entry<DecaFourStateConfig, Set<DecaFourStateConfigPath>> e : pathsPerTgt.entrySet()) {
			Map<Set<DecaStableOutputEvolution>, Set<DecaFourStateConfigPath>> pathsPerEvos = new HashMap<Set<DecaStableOutputEvolution>, Set<DecaFourStateConfigPath>>();
			
			for (DecaFourStateConfigPath p : e.getValue()) {
				Set<DecaStableOutputEvolution> evos = DecaStableOutputEvolution.getOutputEvolutions(p.getCfgs());
				HashMaps.inject(pathsPerEvos, evos, p);
			}
			
			for (Map.Entry<Set<DecaStableOutputEvolution>, Set<DecaFourStateConfigPath>> e2 : pathsPerEvos.entrySet()) {
				DecaFourCoverageStats covStats = new DecaFourCoverageStats();
				
				for (DecaFourStateConfigPath p : e2.getValue()) {
					covStats.addAll(p.getCoverageStats());
				}
				
				result.add(new DecaFourStateConfigPath(e2.getValue().iterator().next(), covStats, PulsePackMap.EMPTY_IN, 0));
			}
		}
		
		return result;
	}
	
	public static Map<JScope, List<DecaFourVertex>> extractPathPerScope(List<DecaFourStateConfig> path) {
		Map<JScope, List<DecaFourVertex>> result = new HashMap<JScope, List<DecaFourVertex>>();
		
		for (Map.Entry<JScope, DecaFourVertex> e : path.get(0).getVtxs().entrySet()) {
			List<DecaFourVertex> p = new ArrayList<DecaFourVertex>();
			p.add(e.getValue());
			result.put(e.getKey(), p);
		}
		
		for (int index = 1; index < path.size(); index++) {
			for (Map.Entry<JScope, DecaFourVertex> e : path.get(index).getVtxs().entrySet()) {
				List<DecaFourVertex> vs = result.get(e.getKey());
				
				if (vs.get(vs.size() - 1) != e.getValue()) {
					vs.add(e.getValue());
				}
			}
		}
		
		return result;
	}
	
	private static boolean isCompatiblePath(Map<JScope, List<DecaFourVertex>> basePathPerScope, Set<DecaStableOutputEvolution> baseEvos, List<DecaFourStateConfig> path) {
		return DecaStableOutputEvolution.getOutputEvolutions(path).equals(baseEvos);
		//return extractPathPerScope(path).equals(basePathPerScope);
	}
	
//	public Set<Map<ReprPort, ASALSymbolicValue>> getSelfLoopZeroPulseSicInputs(Map<ReprPort, ASALSymbolicValue> entryInputs, Collection<ASALPort> timeoutPorts) {
//		Map<JScope, Set<Map<ReprPort, ASALSymbolicValue>>> zeroDiffInputsPerScope = new HashMap<JScope, Set<Map<ReprPort, ASALSymbolicValue>>>();
//		Map<Set<ReprPort>, Map<JScope, Set<PulsePackMap>>> oneDiffInputsPerScope = new HashMap<Set<ReprPort>, Map<JScope, Set<PulsePackMap>>>();
//		
//		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
//			DecaFourVertex succVtx = e1.getValue();
//			
//			for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, Set<InputEquivClz>>> e2 : e1.getValue().getInputsPerNonDetTgtsPerReqSet().entrySet()) {
//				if (e2.getKey().containsMatch(getVtxs())) {
//					for (Map.Entry<DecaFourTgtGrp, Set<InputEquivClz>> e3 : e2.getValue().entrySet()) {
//						if (e3.getKey().getVtxs().contains(succVtx)) {
//							for (InputEquivClz i : e3.getValue()) {
//								for (PulsePackMap inputVal : i.getZeroExternalPulseInputVals()) {
//									InputValDiff diff = getInputValDiff(inputVal, entryInputs);
//									
//									if (diff == NO_DIFF) {
//										//We may combine this with other inputs that DO change:
//										HashMaps.inject(zeroDiffInputsPerScope, e1.getKey(), inputVal);
//									} else {
//										if (diff == TOO_MANY_DIFFS) {
//											//...
//										} else {
//											HashMaps.injectInject(oneDiffInputsPerScope, diff.port.getSyncedPorts(), e1.getKey(), inputVal);
//										}
//									}
//								}
//							}
//						}
//					}
//				}
//			}
//		}
//		
//		Set<Map<ReprPort, ASALSymbolicValue>> result = new HashSet<Map<ReprPort, ASALSymbolicValue>>();
//		
//		for (Map.Entry<Set<ReprPort>, Map<JScope, Set<PulsePackMap>>> e5 : oneDiffInputsPerScope.entrySet()) {
//			Set<JScope> unchangedScopes = new HashSet<JScope>(getVtxs().keySet());
//			unchangedScopes.removeAll(e5.getValue().keySet());
//			
//			if (zeroDiffInputsPerScope.keySet().containsAll(unchangedScopes)) {
//				Map<JScope, Set<Map<ReprPort, ASALSymbolicValue>>> permMap = new HashMap<JScope, Set<Map<ReprPort, ASALSymbolicValue>>>();
//				
//				for (Map.Entry<JScope, Set<Map<ReprPort, ASALSymbolicValue>>> e6 : zeroDiffInputsPerScope.entrySet()) {
//					permMap.put(e6.getKey(), new HashSet<Map<ReprPort, ASALSymbolicValue>>(e6.getValue()));
//				}
//				
//				for (Map.Entry<JScope, Set<PulsePackMap>> e7 : e5.getValue().entrySet()) {
//					HashMaps.injectAll(permMap, e7.getKey(), e7.getValue());
//				}
//				
//				for (Map.Entry<JScope, Set<PulsePackMap>> e8 : e5.getValue().entrySet()) {
//					Map<JScope, Set<Map<ReprPort, ASALSymbolicValue>>> permMapCopy = new HashMap<JScope, Set<Map<ReprPort, ASALSymbolicValue>>>(permMap);
//					permMapCopy.keySet().remove(e8.getKey());
//					
//					for (Map<JScope, Map<ReprPort, ASALSymbolicValue>> perm : HashMaps.allCombinations(permMapCopy)) {
//						Map<ReprPort, ASALSymbolicValue> flatPerm = new HashMap<ReprPort, ASALSymbolicValue>();
//						
//						for (Map.Entry<JScope, Map<ReprPort, ASALSymbolicValue>> e9 : perm.entrySet()) {
//							flatPerm.putAll(e9.getValue());
//						}
//						
//						for (PulsePackMap m : e8.getValue()) {
//							Map<ReprPort, ASALSymbolicValue> flatPermCopy = new HashMap<ReprPort, ASALSymbolicValue>();
//							flatPermCopy.putAll(flatPerm);
//							flatPermCopy.putAll(m);
//							
//							if (isConsistentPerm(flatPermCopy)) {
//								InputValDiff diff = getInputValDiff(flatPermCopy, entryInputs);
//								
//								if (diff != NO_DIFF && diff != TOO_MANY_DIFFS) {
//									result.add(flatPermCopy);
//								}
//							}
//						}
//					}
//				}
//			}
//		}
//		
//		return result;
//	}
	
	private static boolean contains(Collection<DecaFourVertex> vs, String s) {
		for (DecaFourVertex v : vs) {
			if (contains(v, s)) {
				return true;
			}
		}
		
		return false;
	}
	
	private static boolean contains(DecaFourVertex v, String s) {
		return clzSetToStr(v.getSysmlClzs()).contains(s);
	}
	
	private static String clzSetToStr(Set<Class<?>> clzSet) {
		return Texts.concat(clzSet, "+", (c) -> { return c.getSimpleName(); });
	}
	
	/**
	 * This method is inefficient, may need rewriting.
	 */
	public DecaFourCoverageStats getCoverageStatsToSuccsViaInputVal(Set<DecaFourStateConfig> succs, PulsePackMap inputVal) {
		Map<DecaFourStateConfig, DecaFourCoverageStats> z = getSuccsViaInputs(inputVal, inputVal);
		DecaFourCoverageStats result = new DecaFourCoverageStats();
		
		for (DecaFourStateConfig succ : succs) {
			DecaFourCoverageStats zs = z.get(succ);
			
			if (zs == null) {
				System.err.println("inputVal = " + inputVal.toString());
				throw new Error("Should not happen!");
			}
			
			result.addAll(zs);
		}
		
		return result;
	}
	
//	/**
//	 * Returns NULL if there is no such input valuation.
//	 */
	public PulsePackMap getSomeInputToSucc(DecaFourStateConfig succ, boolean mustBeDet) {
		PulsePackMap result = PulsePackMap.EMPTY_IN;
		
		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
			DecaFourVertex succVtx = succ.getVtxs().get(e1.getKey());
			boolean found = false;
			
			for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, DecaFourPulsePackMaps>> e2 : e1.getValue().getExternalInputsPerNonDetTgtsPerReqSet2().entrySet()) {
				if (e2.getKey().containsMatch(getVtxs())) {
					for (Map.Entry<DecaFourTgtGrp, DecaFourPulsePackMaps> z : e2.getValue().entrySet()) {
						if (!mustBeDet || z.getKey().getVtxs().size() == 1) {
							if (z.getKey().getVtxs().contains(succVtx)) {
								result = result.combine(z.getValue().getSomeMap());
								found = true;
								break;
							}
						}
					}
					
					if (found) {
						break;
					}
				}
			}
			
			if (!found) {
				return null;
			}
		}
		
		return result;
	}
	
	/**
	 * Return values are EXTERNAL.
	 */
	public Set<PulsePackMap> getSicInputsToSucc(DecaFourStateMachines sms4, DecaFourStateConfig succ, PulsePackMap entryInputs) {
		Map<JScope, LDD<ReprPort, PulsePack>> zeroDiffLddPerScope = new HashMap<JScope, LDD<ReprPort, PulsePack>>();
		Map<JScope, LDD<ReprPort, PulsePack>> oneDiffLddPerScope = new HashMap<JScope, LDD<ReprPort, PulsePack>>();
		
		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
			DecaFourVertex succVtx = succ.getVtxs().get(e1.getKey());
			LDD<ReprPort, PulsePack> zeroDiffLDD = new LDD<ReprPort, PulsePack>(sms4.smPerScope.get(e1.getKey()).externalInputValEncoding);
			LDD<ReprPort, PulsePack> oneDiffLDD = new LDD<ReprPort, PulsePack>(sms4.smPerScope.get(e1.getKey()).externalInputValEncoding);
			
			for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, DecaFourPulsePackMaps>> e2 : e1.getValue().getExternalInputsPerNonDetTgtsPerReqSet2().entrySet()) {
				if (e2.getKey().containsMatch(getVtxs())) {
					for (Map.Entry<DecaFourTgtGrp, DecaFourPulsePackMaps> e3 : e2.getValue().entrySet()) {
						if (e3.getKey().getVtxs().contains(succVtx)) {
							zeroDiffLDD = zeroDiffLDD.union(e3.getValue().getLDD().diff(entryInputs.getPackPerPort(), 0, 1));
							oneDiffLDD = oneDiffLDD.union(e3.getValue().getLDD().diff(entryInputs.getPackPerPort(), 1));
						}
					}
				}
			}
			
			zeroDiffLddPerScope.put(e1.getKey(), zeroDiffLDD);
			
			if (!oneDiffLDD.isEmpty()) {
				oneDiffLddPerScope.put(e1.getKey(), oneDiffLDD);
			}
		}
		
		Set<PulsePackMap> result = new HashSet<PulsePackMap>();
		
		for (Map.Entry<JScope, LDD<ReprPort, PulsePack>> e : oneDiffLddPerScope.entrySet()) {
			for (Map<ReprPort, PulsePack> oneDiff : e.getValue().getElems()) {
				PulsePackMap entryInputsOneDiff = entryInputs.setAll(oneDiff.values());
				boolean found = true;
				
				for (JScope scope : sms4.orderedScopes) {
					if (scope.equals(e.getKey())) {
						continue;
					}
					
					if (!zeroDiffLddPerScope.get(scope).contains(entryInputsOneDiff.getPackPerPort())) {
						found = false;
						break;
					}
				}
				
				if (found) {
					result.add(entryInputsOneDiff);
				}
			}
		}
		
		return result;
	}
	
	public boolean couldBeStable() {
		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
			boolean found = false;
			
			for (Map.Entry<DecaFourTransitionReqSet, DecaFourPulsePackMaps> e2 : e1.getValue().getSelfLoopInputsPerReqSet2().entrySet()) {
				if (e2.getKey().containsMatch(getVtxs())) {
					found = true;
					break;
				}
			}
			
			if (!found) {
				return false;
			}
		}
		
		return true;
	}
	
	public Set<PulsePackMap> computeStableSelfLoopInputVals() {
//		Map<JScope, Set<PulsePackMap>> selfLoopInputsPerScope = new HashMap<JScope, Set<PulsePackMap>>();
//		
//		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
//			Set<PulsePackMap> selfLoopInputs = new HashSet<PulsePackMap>();
//			
//			for (Map.Entry<DecaFourTransitionReqSet, DecaFourPulsePackMaps> e2 : e1.getValue().getSelfLoopInputsPerReqSet2().entrySet()) {
//				if (e2.getKey().containsMatch(getVtxs())) {
//					selfLoopInputs.addAll(e2.getValue().getMaps());
//				}
//			}
//			
//			if (selfLoopInputs.isEmpty()) {
//				return Collections.emptySet();
//			}
//			
//			selfLoopInputsPerScope.put(e1.getKey(), selfLoopInputs);
//		}
//		
//		Set<PulsePackMap> result = new HashSet<PulsePackMap>();
//		
//		HashMaps.searchCombinations(selfLoopInputsPerScope, (x) -> {
//			result.add(PulsePackMap.merge(x.values(), Dir.IN));
//			return null;
//		});
//		
//		return result;
		return null;
	}
	
//	public Set<PulsePackMap> getMinInputsToSucc(DecaFourStateConfig succ, PulsePackMap entryInputs, int maxDiffCount) {
//		Map<JScope, Set<PulsePackMap>> minDiffInputsPerScope = new HashMap<JScope, Set<PulsePackMap>>();
//		
//		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
//			Set<PulsePackMap> minDiffInputs = new HashSet<PulsePackMap>();
//			int minDiffCount = maxDiffCount;
//			
//			DecaFourVertex succVtx = succ.getVtxs().get(e1.getKey());
//			
//			for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, DecaFourPulsePackMaps>> e2 : e1.getValue().getExternalInputsPerNonDetTgtsPerReqSet2().entrySet()) {
//				if (e2.getKey().containsMatch(getVtxs())) {
//					for (Map.Entry<DecaFourTgtGrp, DecaFourPulsePackMaps> e3 : e2.getValue().entrySet()) {
//						if (e3.getKey().getVtxs().contains(succVtx)) {
//							minDiffCount = e3.getValue().addMinDiffInputs(entryInputs, minDiffInputs, minDiffCount);
//						}
//					}
//				}
//			}
//			
//			if (minDiffInputs.isEmpty()) {
//				return Collections.emptySet();
////				throw new Error("Should not happen!");
//			}
//			
//			minDiffInputsPerScope.put(e1.getKey(), minDiffInputs);
//		}
//		
//		Set<PulsePackMap> result = new HashSet<PulsePackMap>();
//		
//		//Note: it can be that synced ports prevent a suitable (minimal) combination!
//		HashMaps.searchCombinations(minDiffInputsPerScope, (x) -> {
//			result.add(PulsePackMap.merge(x.values(), Dir.IN));
//			return null;
//		});
//		
//		return result;
//	}
	
//	public Set<PulsePackMap> getMinInputsToMissingTrs(Set<DecaFourTransition> beenHere, PulsePackMap entryInputs) {
//		Map<JScope, Set<PulsePackMap>> minDiffInputsPerScope = new HashMap<JScope, Set<PulsePackMap>>();
//		
//		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
//			Set<PulsePackMap> minDiffInputs = new HashSet<PulsePackMap>();
//			int minDiffCount = 2;
//			
////			DecaFourVertex succVtx = succ.getVtxs().get(e1.getKey());
//			
//			for (DecaFourTransition t : e1.getValue().getOutgoing()) {
//				if (!beenHere.contains(t)) {
//					if (t.getReqs().containsMatch(getVtxs())) {
//						minDiffCount = t.getInputs2().addMinDiffInputs(entryInputs, minDiffInputs, minDiffCount);
////						e1.getValue().getExternalInputsPerNonDetTgtsPerReqSet2().get(t.getReqs())
//					}
//				}
//			}
//			
//			if (minDiffInputs.isEmpty()) {
//				for (DecaFourTransition t : e1.getValue().getOutgoing()) {
//					if (t.getReqs().containsMatch(getVtxs())) {
//						minDiffCount = t.getInputs2().addMinDiffInputs(entryInputs, minDiffInputs, minDiffCount);
//					}
//				}
//			}
//			
//			minDiffInputsPerScope.put(e1.getKey(), minDiffInputs);
//		}
//		
//		Set<PulsePackMap> result = new HashSet<PulsePackMap>();
//		
//		//Note: it can be that synced ports prevent a suitable (minimal) combination!
//		HashMaps.searchCombinations(minDiffInputsPerScope, (x) -> {
//			result.add(PulsePackMap.merge(x.values(), Dir.IN));
//			return null;
//		});
//		
//		return result;
//	}
	
//	private static boolean isConsistentPerm(Map<ReprPort, ASALSymbolicValue> flatPerm) {
//		for (Map.Entry<ReprPort, ASALSymbolicValue> e : flatPerm.entrySet()) {
//			if (JPulse.class.isAssignableFrom(e.getKey().getType())) {
//				if (e.getValue().toBoolean()) {
//					for (ReprPort sp : e.getKey().getSyncedPorts()) {
//						ASALSymbolicValue v = flatPerm.get(sp);
//						
//						if (v == null || !v.toBoolean()) {
//							return false;
//						}
//						
//						for (ReprPort rdp : e.getKey().getDataPorts()) {
//							for (ReprPort srdp : rdp.getSyncedPorts()) {
//								if (!flatPerm.get(srdp).equals(flatPerm.get(rdp))) {
//									return false;
//								}
//							}
//						}
//					}
//				} else {
//					for (ReprPort sp : e.getKey().getSyncedPorts()) {
//						if (flatPerm.get(sp).toBoolean()) {
//							return false;
//						}
//					}
//				}
//			} else {
//				if (e.getKey().getPulsePort() == null) {
//					for (ReprPort sp : e.getKey().getSyncedPorts()) {
//						if (!flatPerm.get(sp).equals(e.getValue())) {
//							return false;
//						}
//					}
//				}
//			}
//		}
//		
//		return true;
//	}
//	
//	private static class InputValDiff {
//		public final ReprPort port;
//		
//		public InputValDiff(ReprPort port) {
//			this.port = port;
//		}
//	}
//	
//	private final static InputValDiff NO_DIFF = new InputValDiff(null);
//	private final static InputValDiff TOO_MANY_DIFFS = new InputValDiff(null);
//	
//	private static InputValDiff getInputValDiff(Map<ReprPort, ASALSymbolicValue> inputVal, Map<ReprPort, ASALSymbolicValue> entryInputVal) {
//		InputValDiff result = NO_DIFF;
//		
//		for (Map.Entry<ReprPort, ASALSymbolicValue> e : inputVal.entrySet()) {
//			if (e.getKey().isPortToEnvironment()) {
//				if (JPulse.class.isAssignableFrom(e.getKey().getType())) {
//					if (e.getValue().toBoolean()) {
//						if (result.port != null) {
//							if (result.port != e.getKey().getAdapterLabelPort()) {
//								return TOO_MANY_DIFFS;
//							}
//						} else {
//							result = new InputValDiff(e.getKey().getAdapterLabelPort());
//						}
//					}
//				} else {
//					if (e.getKey().getPulsePort() == null) {
//						if (!e.getValue().equals(entryInputVal.get(e.getKey()))) {
//							if (result.port != null) {
//								if (result.port != e.getKey().getAdapterLabelPort()) {
//									return TOO_MANY_DIFFS;
//								}
//							} else {
//								result = new InputValDiff(e.getKey().getAdapterLabelPort());
//							}
//						}
//					}
//				}
//			}
//		}
//		
//		return result;
//	}
	
//	public Map<JScope, Set<InputEquivClz>> computeEnabledPerScope() {
//		Map<JScope, Set<InputEquivClz>> enabledPerScope = new HashMap<JScope, Set<InputEquivClz>>();
//		
//		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
//			Set<InputEquivClz> enabled = new HashSet<InputEquivClz>();
//			
//			for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, Set<PulsePackMap>>> e2 : e1.getValue().getInputsPerNonDetTgtsPerReqSet().entrySet()) {
//				if (e2.getKey().containsMatch(getVtxs())) {
//					for (Map.Entry<DecaFourTgtGrp, Set<PulsePackMap>> e3 : e2.getValue().entrySet()) {
//						enabled.addAll(e3.getValue());
//					}
//				}
//			}
//			
//			enabledPerScope.put(e1.getKey(), enabled);
//		}
//		
//		return enabledPerScope;
//	}
	
//	public Map<JScope, Map<DecaFourVertex, Set<InputEquivClz>>> computeInputsPerSuccsPerScope() {
//		Map<JScope, Map<DecaFourVertex, Set<InputEquivClz>>> allTgtsPerScope = new HashMap<JScope, Map<DecaFourVertex, Set<InputEquivClz>>>();
//		
//		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
//			Map<DecaFourVertex, Set<InputEquivClz>> allTgts = new HashMap<DecaFourVertex, Set<InputEquivClz>>();
//			
//			for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, Set<InputEquivClz>>> e2 : e1.getValue().getInputsPerNonDetTgtsPerReqSet().entrySet()) {
//				if (e2.getKey().containsMatch(getVtxs())) {
//					for (Map.Entry<DecaFourTgtGrp, Set<InputEquivClz>> e3 : e2.getValue().entrySet()) {
//						for (DecaFourVertex vtx : e3.getKey().getVtxs()) {
//							HashMaps.injectAll(allTgts, vtx, e3.getValue());
//						}
//					}
//				}
//			}
//			
//			allTgtsPerScope.put(e1.getKey(), allTgts);
//		}
//		
//		return allTgtsPerScope;
//	}
	
	public Map<JScope, Set<DecaFourVertex>> computeSuccsPerScope() {
		Map<JScope, Set<DecaFourVertex>> allTgtsPerScope = new HashMap<JScope, Set<DecaFourVertex>>();
		
		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
			Set<DecaFourVertex> allTgts = new HashSet<DecaFourVertex>();
			
			for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, Set<PulsePackMap>>> e2 : e1.getValue().getInputsPerNonDetTgtsPerReqSet().entrySet()) {
				if (e2.getKey().containsMatch(getVtxs())) {
					for (DecaFourTgtGrp vtxs : e2.getValue().keySet()) {
						allTgts.addAll(vtxs.getVtxs());
					}
				}
			}
			
			allTgtsPerScope.put(e1.getKey(), allTgts);
		}
		
		return allTgtsPerScope;
	}
	
	public Map<JScope, List<DecaFourVertex>> computeOrderedSuccsPerScope() {
		Map<JScope, List<DecaFourVertex>> result = new HashMap<JScope, List<DecaFourVertex>>();
		
		for (Map.Entry<JScope, Set<DecaFourVertex>> e : computeSuccsPerScope().entrySet()) {
			List<DecaFourVertex> orderedVtxs = new ArrayList<DecaFourVertex>(e.getValue());
			Collections.sort(orderedVtxs);
			result.put(e.getKey(), orderedVtxs);
		}
		
		return result;
	}
	
	public Set<DecaFourStateConfig> computeSuccs() {
		return computeSuccs(computeSuccsPerScope());
	}
	
	public Set<DecaFourStateConfig> computeSuccs(Map<JScope, ? extends Collection<DecaFourVertex>> succsPerScope) {
		Set<DecaFourStateConfig> result = new HashSet<DecaFourStateConfig>();
		
		HashMaps.searchCombinations(succsPerScope, (perm) -> {
			DecaFourStateConfig f2 = new DecaFourStateConfig(perm);
			
			if (result.add(f2)) {
				f2.level = level + 1;
			}
			
			return null;
		});
		
		return result;
	}
	
	public boolean isNonDetInputVal(PulsePackMap inputVal) {
		Set<DecaFourStateConfig> result = new HashSet<DecaFourStateConfig>();
		result.add(this);
		
		Set<DecaFourStateConfig> fringe = new HashSet<DecaFourStateConfig>();
		Set<DecaFourStateConfig> newFringe = new HashSet<DecaFourStateConfig>();
		fringe.add(this);
		
		while (fringe.size() > 0) {
			newFringe.clear();
			
			for (DecaFourStateConfig cfg : fringe) {
				for (DecaFourStateConfig succCfg : cfg.computeSuccsViaInputVal(inputVal).keySet()) {
					if (result.add(succCfg)) {
						newFringe.add(succCfg);
					}
				}
			}
			
			fringe.clear();
			fringe.addAll(newFringe);
			
//			System.out.println("[" + LocalTime.now() + "] #result = " + result.size() + " (+" + fringe.size() + ")");
			
			if (result.size() > 1) {
				return true;
			}
		}
		
		return false;
	}
	
	public Set<DecaFourStateConfig> computeTransitiveSuccsViaInputVal(PulsePackMap inputVal) {
		Set<DecaFourStateConfig> result = new HashSet<DecaFourStateConfig>();
		result.add(this);
		
		Set<DecaFourStateConfig> fringe = new HashSet<DecaFourStateConfig>();
		Set<DecaFourStateConfig> newFringe = new HashSet<DecaFourStateConfig>();
		fringe.add(this);
		
		while (fringe.size() > 0) {
			newFringe.clear();
			
			for (DecaFourStateConfig cfg : fringe) {
				for (DecaFourStateConfig succCfg : cfg.computeSuccsViaInputVal(inputVal).keySet()) {
					if (result.add(succCfg)) {
						newFringe.add(succCfg);
					}
				}
			}
			
			fringe.clear();
			fringe.addAll(newFringe);
			
//			System.out.println("[" + LocalTime.now() + "] #result = " + result.size() + " (+" + fringe.size() + ")");
		}
		
		return result;
	}
	
	/**
	 * Input valuation can be partial.
	 */
	public Map<DecaFourStateConfig, DecaFourCoverageStats> computeSuccsViaInputVal(PulsePackMap inputVal) {
		return computeSuccsViaInputValB(inputVal, false);
	}
	
	public Map<DecaFourStateConfig, DecaFourCoverageStats> computeSuccsViaInputValB(PulsePackMap inputVal, boolean limit) {
//		Map<DecaFourVertex, Set<DecaFourTransition>> trsPerTgt = new HashMap<DecaFourVertex, Set<DecaFourTransition>>();
//		Map<JScope, Set<DecaFourVertex>> allTgtsPerScope = new HashMap<JScope, Set<DecaFourVertex>>();
//		
//		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
//			PulsePackMap premise = inputVal.extractOwnerMap(e1.getKey());
//			Set<DecaFourVertex> allTgts = new HashSet<DecaFourVertex>();
//			
//			for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, DecaFourPulsePackMaps>> e2 : e1.getValue().getExternalInputsPerNonDetTgtsPerReqSet2().entrySet()) {
//				if (e2.getKey().containsMatch(getVtxs())) {
//					for (Map.Entry<DecaFourTgtGrp, DecaFourPulsePackMaps> z : e2.getValue().entrySet()) {
//						if (z.getValue().couldBeImpliedBy(premise)) {
//							allTgts.addAll(z.getKey().getVtxs());
//							
//							trsPerTgt;
//						}
//					}
//				}
//			}
//			
//			//Because input-enabled:
//			if (allTgts.isEmpty()) {
//				throw new Error("Should not happen!");
//			}
//			
//			allTgtsPerScope.put(e1.getKey(), allTgts);
//		}
//		
//		Map<DecaFourStateConfig, DecaFourCoverageStats> result = new HashMap<DecaFourStateConfig, DecaFourCoverageStats>();
//		
//		if (!limit || HashMaps.getCombinationCount(allTgtsPerScope, false) <= 100) {
//			for (Map<JScope, DecaFourVertex> perm : HashMaps.allCombinations(allTgtsPerScope)) {
//				result.add(new DecaFourStateConfig(perm));
//			}
//		}
//		
//		return result;
		
		Map<DecaFourVertex, Set<DecaFourTransition>> trsPerTgt = new HashMap<DecaFourVertex, Set<DecaFourTransition>>();
		Map<JScope, Set<DecaFourVertex>> pssPerScope = new HashMap<JScope, Set<DecaFourVertex>>(); //Possible SuccessorS
		int index = 0;
		
		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
			Set<DecaFourVertex> pss = new HashSet<DecaFourVertex>();
			
			for (Map.Entry<DecaFourTransitionReqSet, Map<PulsePackMap, Set<DecaFourTransition>>> e2 : e1.getValue().getTrsPerNonDetTgtsPerReqSet().entrySet()) {
				if (e2.getKey().containsMatch(getVtxs())) {
//					System.out.println("option " + index);
					index++;
					
					for (Map.Entry<PulsePackMap, Set<DecaFourTransition>> e3 : e2.getValue().entrySet()) {
						if (inputVal.implies(e3.getKey())) {
							for (DecaFourTransition t : e3.getValue()) {
								for (DecaFourVertex tgt : t.getTgtGrp().getVtxs()) {
									HashMaps.inject(trsPerTgt, tgt, t);
								}
								
								pss.addAll(t.getTgtGrp().getVtxs());
							}
							
//							protoTrsDest.addAll(e1.getValue().getProtoTrsPerReqSet().get(
//							System.out.println("adding " + e3.getKey().getVtxs().size());
						}
					}
				}
			}
			
			if (pss.isEmpty()) {
				throw new Error("Should not happen!");
			}
			
			pssPerScope.put(e1.getKey(), pss);
		}
		
//		System.out.println("#combis = " + HashMaps.getCombinationCount(pssPerScope, false));
		
		Map<DecaFourStateConfig, DecaFourCoverageStats> result = new HashMap<DecaFourStateConfig, DecaFourCoverageStats>();
		
		for (Map<JScope, DecaFourVertex> perm : HashMaps.allCombinations(pssPerScope)) {
			DecaFourCoverageStats covStats = new DecaFourCoverageStats();
			
			for (DecaFourVertex tgt : perm.values()) {
				covStats.add4(trsPerTgt.get(tgt));
			}
			
			result.put(new DecaFourStateConfig(perm), covStats);
		}
		
		return result;
	}
	
//	/**
//	 * Fixed point of successors after a specific input valuation.
//	 * Pulses are deactivated after the first step.
//	 */
//	public Set<DecaFourStateConfig> computeReachableViaInputVal(PulsePackMap inputVal) {
//		Set<DecaFourStateConfig> firstFringe = computeSuccsViaInputVal(inputVal);
//		PulsePackMap deactivatedInput = inputVal.deactivate();
//		
//		Set<DecaFourStateConfig> result = new HashSet<DecaFourStateConfig>();
//		result.addAll(firstFringe);
//		
//		Set<DecaFourStateConfig> fringe = new HashSet<DecaFourStateConfig>();
//		Set<DecaFourStateConfig> newFringe = new HashSet<DecaFourStateConfig>();
//		fringe.addAll(firstFringe);
//		
//		while (fringe.size() > 0) {
//			newFringe.clear();
//			
//			for (DecaFourStateConfig cfg : fringe) {
//				for (DecaFourStateConfig succ : cfg.computeSuccsViaInputVal(deactivatedInput)) {
//					if (result.add(succ)) {
//						newFringe.add(succ);
//					}
//				}
//			}
//			
//			fringe.clear();
//			fringe.addAll(newFringe);
//		}
//		
//		return result;
//	}
	
	/**
	 * Adds to the given set the successors of this state configuration.
	 */
	public long addAllSuccs(NormMap<DecaFourStateConfig> dest, NormMap<Pair<DecaFourVertex>> pairs, DecaFourCoverageStats coverageStatsDest) {
		Map<JScope, Set<DecaFourVertex>> allTgtsPerScope = new HashMap<JScope, Set<DecaFourVertex>>();
		
		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
			Set<DecaFourVertex> allTgts = new HashSet<DecaFourVertex>();
			
			int b1 = 0;
			
			for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, Set<PulsePackMap>>> e2 : e1.getValue().getInputsPerNonDetTgtsPerReqSet().entrySet()) {
				if (e2.getKey().containsMatch(getVtxs())) {
					b1++;
					
					for (DecaFourTgtGrp vtxs : e2.getValue().keySet()) {
						allTgts.addAll(vtxs.getVtxs());
					}
					
					for (Set<DecaFourTransition> trs : e1.getValue().getTrsPerNonDetTgtsPerReqSet().get(e2.getKey()).values()) {
//						for (DecaFourTransition t : trs) {
//							proto.addAll(t.getProtoTrs());
//						}
						coverageStatsDest.add4(trs);
					}
//					Set<ProtoTransition> protoTrs = ;
//					
//					if (protoTrs != null) {
//						proto.addAll(protoTrs);
//					}
				}
			}
			
			if (allTgts.isEmpty()) {
				System.out.println("No targets for " + e1.getKey().getName() + " -> " + e1.getValue().getName());
				System.out.println("e1.getValue() = " + e1.getValue().getInputsPerNonDetTgtsPerReqSet().size());
				System.out.println("e2.getKey().containsMatch(getVtxs()) = " + b1);
				
//				for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, Set<PulsePackMap>>> e2 : e1.getValue().getInputsPerNonDetTgtsPerReqSet().entrySet()) {
//					System.out.println("() " + e2.getKey().containsMatch(getVtxs()) + " =>");
//					
//					for (Map<JScope, DecaFourVertex> x : e2.getKey().getReqCfgs()) {
//						List<String> elems = new ArrayList<String>();
//						
//						for (Map.Entry<JScope, DecaFourVertex> xx : x.entrySet()) {
//							elems.add(xx.getValue().getName());
//						}
//						
//						System.out.println("\t" + Texts.concat(elems, " || "));
//					}
//				}
				
				//fp[0][WATING]$0
				//est[0][NO_OPERATING_VOLTAGE]$0
				//sec[0][NOT_READY_FOR_CONNECTION]$0
				System.out.println("b/c we are at " + getDescription());
				
				throw new Error("Should not happen!");
			}
			
			allTgtsPerScope.put(e1.getKey(), allTgts);
		}
		
		HashMaps.searchCombinations(allTgtsPerScope, (perm) -> {
			DecaFourStateConfig f2 = new DecaFourStateConfig(perm);
			pairs.addAll(f2.computeVtxPairsFrom(this));
			
			if (dest.add(f2)) {
				f2.level = level + 1;
			}
			
			return null;
		});
		
		return HashMaps.getCombinationCount(allTgtsPerScope, false);
	}
	
	public long addAllDetSuccs(NormMap<DecaFourStateConfig> dest, NormMap<Pair<DecaFourVertex>> pairs, DecaFourCoverageStats coverageStatsDest) {
		Map<JScope, Set<DecaFourVertex>> allTgtsPerScope = new HashMap<JScope, Set<DecaFourVertex>>();
		
		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
			Set<DecaFourVertex> allTgts = new HashSet<DecaFourVertex>();
			
			for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, Set<PulsePackMap>>> e2 : e1.getValue().getInputsPerDetTgtsPerReqSet().entrySet()) {
				if (e2.getKey().containsMatch(getVtxs())) {
					for (DecaFourTgtGrp vtxs : e2.getValue().keySet()) {
						allTgts.addAll(vtxs.getVtxs());
					}
				}
			}
			
			if (allTgts.isEmpty()) {
				return 0L;
			}
			
			allTgtsPerScope.put(e1.getKey(), allTgts);
		}
		
		HashMaps.searchCombinations(allTgtsPerScope, (perm) -> {
			DecaFourStateConfig f2 = new DecaFourStateConfig(perm);
			pairs.addAll(f2.computeVtxPairsFrom(this));
			
			if (dest.add(f2)) {
				f2.level = level + 1;
			}
			
			return null;
		});
		
		return HashMaps.getCombinationCount(allTgtsPerScope, false);
	}
	
	public TransitiveLDD<JScope, DecaFourVertex> addAllSuccs2(DecaFourStateMachines sms4, NormMap<DecaFourStateConfig> dest, NormMap<Pair<DecaFourVertex>> pairs, DecaFourCoverageStats coverageStatsDest) {
		Map<JScope, Set<DecaFourVertex>> allTgtsPerScope = new HashMap<JScope, Set<DecaFourVertex>>();
		
		for (Map.Entry<JScope, DecaFourVertex> e1 : getVtxs().entrySet()) {
			Set<DecaFourVertex> allTgts = new HashSet<DecaFourVertex>();
			
			int b1 = 0;
			
			for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, Set<PulsePackMap>>> e2 : e1.getValue().getInputsPerNonDetTgtsPerReqSet().entrySet()) {
				if (e2.getKey().containsMatch(getVtxs())) {
					b1++;
					
					for (DecaFourTgtGrp vtxs : e2.getValue().keySet()) {
						allTgts.addAll(vtxs.getVtxs());
					}
					
					for (Set<DecaFourTransition> trs : e1.getValue().getTrsPerNonDetTgtsPerReqSet().get(e2.getKey()).values()) {
						coverageStatsDest.add4(trs);
					}
				}
			}
			
			if (allTgts.isEmpty()) {
				System.out.println("No targets for " + e1.getKey().getName() + " -> " + e1.getValue().getName());
				System.out.println("e1.getValue() = " + e1.getValue().getInputsPerNonDetTgtsPerReqSet().size());
				System.out.println("e2.getKey().containsMatch(getVtxs()) = " + b1);
				
//				for (Map.Entry<DecaFourTransitionReqSet, Map<DecaFourTgtGrp, Set<PulsePackMap>>> e2 : e1.getValue().getInputsPerNonDetTgtsPerReqSet().entrySet()) {
//					System.out.println("() " + e2.getKey().containsMatch(getVtxs()) + " =>");
//					
//					for (Map<JScope, DecaFourVertex> x : e2.getKey().getReqCfgs()) {
//						List<String> elems = new ArrayList<String>();
//						
//						for (Map.Entry<JScope, DecaFourVertex> xx : x.entrySet()) {
//							elems.add(xx.getValue().getName());
//						}
//						
//						System.out.println("\t" + Texts.concat(elems, " || "));
//					}
//				}
				
				//fp[0][WATING]$0
				//est[0][NO_OPERATING_VOLTAGE]$0
				//sec[0][NOT_READY_FOR_CONNECTION]$0
				System.out.println("b/c we are at " + getDescription());
				
				throw new Error("Should not happen!");
			}
			
			allTgtsPerScope.put(e1.getKey(), allTgts);
		}
		
		TransitiveLDD<JScope, DecaFourVertex> result = new TransitiveLDD<JScope, DecaFourVertex>(sms4.vtxsPerScope, sms4.orderedScopes);
		
		HashMaps.searchCombinations(allTgtsPerScope, (perm) -> {
			DecaFourStateConfig f2 = new DecaFourStateConfig(perm);
			pairs.addAll(f2.computeVtxPairsFrom(this));
			
			result.add(getVtxs(), f2.getVtxs());
			
			if (dest.add(f2)) {
				f2.level = level + 1;
			}
			
			return null;
		});
		
		return result;
	}
	
	public Set<Pair<DecaFourVertex>> computeVtxPairsFrom(DecaFourStateConfig src) {
		Set<Pair<DecaFourVertex>> result = new HashSet<Pair<DecaFourVertex>>();
		
		for (Map.Entry<JScope, DecaFourVertex> e : src.vtxs.entrySet()) {
			result.add(new Pair<DecaFourVertex>(e.getValue(), vtxs.get(e.getKey())));
		}
		
		return result;
	}
	
	@Override
	public Set<LineageObject> getImpliedElements() {
		Set<LineageObject> result = new HashSet<LineageObject>();
		result.addAll(getVtxs().values());
		return result;
	}
	
	@Override
	public Set<LineageObject> getInferredElements() {
		return Collections.emptySet();
	}
	
	@Override
	public int hashCode() {
		return vtxsHashCode;
	}
	
	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!(obj instanceof DecaFourStateConfig)) {
			return false;
		}
		DecaFourStateConfig other = (DecaFourStateConfig) obj;
		return Objects.equals(vtxs, other.vtxs);
	}
	
	public List<DecaFourVertex> getOrderedVtxs() {
		SortedMap<String, JScope> orderedScopes = new TreeMap<String, JScope>();
		
		for (JScope scope : getVtxs().keySet()) {
			orderedScopes.put(scope.getName(), scope);
		}
		
		List<DecaFourVertex> result = new ArrayList<DecaFourVertex>(); 
		
		for (Map.Entry<String, JScope> e : orderedScopes.entrySet()) {
			result.add(getVtxs().get(e.getValue()));
		}
		
		return result;
	}
	
	public String getDescription() {
		//return Texts.concat(getOrderedVtxs(), "+", (v) -> { return v.getName() + "$" + v.getId(); });
		return Texts.concat(getOrderedVtxs(), "+", (v) -> { return v.getName(); }) + "::" + vtxsHashCode;
	}
	
	public String getDescription(String lineSep) {
//		return Texts.concat(getOrderedVtxs(), lineSep, (v) -> { return v.getName() + "$" + v.getId(); });
		return Texts.concat(getOrderedVtxs(), lineSep, (v) -> { return v.getName(); });
	}
	
	@Override
	public String toString() {
		return getDescription();
	}
}

