package jlx.behave.proto;

import java.io.*;
import java.time.LocalTime;
import java.util.*;

import jlx.asal.j.*;
import jlx.asal.rewrite.ASALSymbolicValue;
import jlx.asal.vars.ASALPort;
import jlx.models.UnifyingBlock.ReprPort;
import jlx.utils.*;
import jlx.utils.ldd.*;

public class DecaFourStateMachines implements LineageObject {
	public final String name;
	public final Map<JScope, DecaFourStateMachine> smPerScope;
	public final Map<JScope, DecaFourVertex> initialVertices;
	public final Map<JScope, Set<DecaFourVertex>> vtxsPerScope;
	public final LDDEncoding<JScope, DecaFourVertex> cfgEncoding;
	public final PulsePackMap initialInputs;
	public final Set<DecaFourVertex> vertices;
	public final Set<DecaFourTransition> transitions;
	public final Set<ProtoTransition> initProtoTransitions;
	public final Set<InputEquivClz> inputEquivClzs;
	public final DecaFourStateConfig initCfg;
	public final NormMap<DecaFourStateConfig> configs;
	public final Set<DecaFourStateConfig> potStableCfgs;
	public final NormMap<Pair<DecaFourVertex>> vtxPairs;
	public final DecaFourCoverageStats coverageStats;
	public final NormMap<DecaFourVertex> vtxs;
	public final Map<ASALPort, ASALPort> timeoutPortPerDurationPort;
	public final List<JScope> orderedScopes;
	public final Map<DecaFourStateConfig, Integer> levelPerCfg;
	public final Map<Integer, Set<DecaFourStateConfig>> cfgsPerLevel;
	
	private double totalTrsCount;
	private double totalDetTrsCount;
	
	private static class Worker extends ConcurrentWorker {
		private final NormMap<DecaFourStateConfig> elems1;
		private final NormMap<Pair<DecaFourVertex>> elems2;
		private final DecaFourCoverageStats elems3;
		private double trsCount;
		private double detTrsCount;
		private TransitiveLDD<JScope, DecaFourVertex> trldd;
		
		public Worker(DecaFourStateMachines sms4) {
			elems1 = new NormMap<DecaFourStateConfig>();
			elems2 = new NormMap<Pair<DecaFourVertex>>();
			elems3 = new DecaFourCoverageStats();
			trsCount = 0;
			detTrsCount = 0f;
			
			trldd = new TransitiveLDD<JScope, DecaFourVertex>(sms4.vtxsPerScope, sms4.orderedScopes);
		}
		
		@Override
		public String getSuffix() {
			return "; #cfgs = " + elems1.size() + "; #pairs = " + elems2.size() + "; #proto = " + elems3.toString() + "; #trs = " + trsCount + "; #det-trs = " + detTrsCount;
		}
		
		@Override
		public String toString() {
			return getSuffix();
		}
	}
	
	public void initCfgs() {
		System.out.println("[cfgs][" + LocalTime.now() + "] #cfgs = 0");
		
		configs.clear();
		configs.add(initCfg);
		vtxPairs.clear();
		coverageStats.clear();
		
		totalTrsCount = 0;
		totalDetTrsCount = 0;
		levelPerCfg.clear();
		cfgsPerLevel.clear();
		int level = 0;
		
		TransitiveLDD<JScope, DecaFourVertex> trldd = new TransitiveLDD<JScope, DecaFourVertex>(vtxsPerScope, orderedScopes);
		
		Set<DecaFourStateConfig> fringe = new HashSet<DecaFourStateConfig>();
		fringe.add(initCfg);
		
		while (fringe.size() > 0) {
			Set<DecaFourStateConfig> levelCfgs = new HashSet<DecaFourStateConfig>();
			
			for (DecaFourStateConfig cfg : fringe) {
				levelPerCfg.put(cfg, level);
				levelCfgs.add(cfg);
			}
			
			cfgsPerLevel.put(level, levelCfgs);
			level++;
			
			ConcurrentWork<DecaFourStateConfig, Worker> cw = new ConcurrentWork<DecaFourStateConfig, Worker>();
			
			for (int index = 1; index <= 4; index++) {
				cw.getWorkers().add(new Worker(this));
			}
			
			cw.apply(fringe, (cfg, worker) -> {
				worker.trsCount += cfg.addAllSuccs(worker.elems1, worker.elems2, worker.elems3);
//				worker.detTrsCount += cfg.addAllDetSuccs(worker.elems1, worker.elems2, worker.elems3);
//				worker.trldd = worker.trldd.add(cfg.addAllSuccs2(this, worker.elems1, worker.elems2, worker.elems3));
			});
			
			fringe.clear();
			
			for (Worker worker : cw.getWorkers()) {
				for (DecaFourStateConfig cfg : worker.elems1) {
					if (configs.add(cfg)) {
						cfg.level = level;
						fringe.add(cfg);
					}
				}
				
				vtxPairs.addAll(worker.elems2);
				trldd = trldd.add(worker.trldd);
				totalTrsCount += worker.trsCount;
				totalDetTrsCount += worker.detTrsCount;
				
				coverageStats.addAll(worker.elems3);
			}
			
			System.out.println("[cfgs][" + LocalTime.now() + "] #cfgs = " + configs.size() + " (+" + fringe.size() + "); #pairs = " + vtxPairs.size() + "; #proto = " + coverageStats.toString() + "; #trs = " + totalTrsCount + "; #det-trs = " + totalDetTrsCount);
		}
		
//		CLI.waitForEnter();
	}
	
	public DecaFourStateMachines(String name, Collection<DecaTritoStateMachine> sources, boolean explore) {
		this.name = name;
		
		Map<JScope, DecaTritoStateMachine> oldSmPerScope = new HashMap<JScope, DecaTritoStateMachine>();
		Map<DecaTritoVertex, DecaFourVertex> newVertexPerOldVertex = new HashMap<DecaTritoVertex, DecaFourVertex>();
		vtxsPerScope = new HashMap<JScope, Set<DecaFourVertex>>();
		
		for (DecaTritoStateMachine s : sources) {
			oldSmPerScope.put(s.scope, s);
			Set<DecaFourVertex> vtxs = new HashSet<DecaFourVertex>();
			
			for (DecaTritoVertex oldVertex : s.vertices.values()) {
				DecaFourVertex newVertex = new DecaFourVertex(s.scope, oldVertex);
				newVertexPerOldVertex.put(oldVertex, newVertex);
				vtxs.add(newVertex);
			}
			
			vtxsPerScope.put(s.scope, vtxs);
		}
		
		cfgEncoding = new LDDEncoding<JScope, DecaFourVertex>(vtxsPerScope);
		
		smPerScope = new HashMap<JScope, DecaFourStateMachine>();
		initialVertices = new HashMap<JScope, DecaFourVertex>();
		Map<ReprPort, PulsePack> initialValuePerPort = new HashMap<ReprPort, PulsePack>();
		vertices = new HashSet<DecaFourVertex>();
		transitions = new HashSet<DecaFourTransition>();
		initProtoTransitions = new HashSet<ProtoTransition>();
		inputEquivClzs = new HashSet<InputEquivClz>();
		timeoutPortPerDurationPort = new HashMap<ASALPort, ASALPort>();
		cfgsPerLevel = new HashMap<Integer, Set<DecaFourStateConfig>>();
		levelPerCfg = new HashMap<DecaFourStateConfig, Integer>();
		
		int reqSetCount = 0;
		
		int multiValueInputPortCount = 0;
		int constantInputPortCount = 0;
		int internalInputPortCount = 0;
		int internalOutputPortCount = 0;
		double inputLabelCount = 1.0;
		int potentialOutputPortCount = 0;
		
		for (DecaTritoStateMachine s : sources) {
			System.out.println("[" + LocalTime.now() + "] Building state machine for " + s.scope.getName());
			DecaFourStateMachine newSm = new DecaFourStateMachine(s, oldSmPerScope, newVertexPerOldVertex, cfgEncoding);
			smPerScope.put(s.scope, newSm);
			initialVertices.put(s.scope, newSm.initialVertex);
			initialValuePerPort.putAll(newSm.initialInputs.getPackPerPort());
			vertices.addAll(newSm.vertices);
			vtxsPerScope.put(s.scope, newSm.vertices);
			transitions.addAll(newSm.transitions);
			initProtoTransitions.addAll(newSm.initProtoTransitions);
			inputEquivClzs.addAll(newSm.inputEquivClzs);
			timeoutPortPerDurationPort.putAll(s.timeoutPortPerDurationPort);
			
			for (DecaFourVertex v : newSm.vertices) {
				reqSetCount += v.getNondetOutgoingPerReqSet().size();
			}
			
			for (Map.Entry<ReprPort, Set<PulsePack>> e : s.pulsePacksPerInputPort.entrySet()) {
				if (e.getKey().isPortToEnvironment()) {
					if (e.getKey().getDir() == Dir.IN) { //Should always hold...
						if (e.getValue().size() > 1) {
							multiValueInputPortCount++;
							inputLabelCount *= e.getValue().size();
						} else {
							constantInputPortCount++;
						}
					}
				} else {
					internalInputPortCount++;
				}
			}
			
			for (ReprPort ep : PulsePackMap.from(s.initialVertex.getOutputVal(), Dir.OUT).getPackPerPort().keySet()) {
				if (ep.isPortToEnvironment()) {
					potentialOutputPortCount++;
				} else {
					internalOutputPortCount++;
				}
			}
			
		}
		
		System.out.println("#reqSets = " + reqSetCount);
		
		System.out.println("#multi-value-env-input-ports = " + multiValueInputPortCount);
		System.out.println("#constant-env-input-ports = " + constantInputPortCount);
		System.out.println("#env-input-labels = " + inputLabelCount);
		System.out.println("#potential-env-output-ports = " + potentialOutputPortCount);
		System.out.println("#internal-input-ports = " + internalInputPortCount);
		System.out.println("#internal-output-ports = " + internalOutputPortCount);
		
		String[][] smdTable = new String[7][11];
		smdTable[0][0] = "";
		smdTable[0][1] = "Step 0";
		smdTable[0][2] = "Step 1";
		smdTable[0][3] = "Step 2";
		smdTable[0][4] = "Step 3";
		smdTable[0][5] = "Step 4, 5";
		smdTable[0][6] = "Step 6";
		smdTable[0][7] = "Step 7";
		smdTable[0][8] = "Step 8, 9, 10";
		smdTable[0][9] = "Instantiated";
		smdTable[0][10] = "DecaFour";
		
		int index = 1;
		
		for (String x : new String[] { "prim", "sec", "est", "sp", "fp", "p3" }) {
			for (Map.Entry<JScope, DecaFourStateMachine> e : smPerScope.entrySet()) {
				if (e.getKey().getName().equals(x)) {
					smdTable[index][0] = x;
					smdTable[index][1] = "" + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.vertices.size();
					smdTable[index][2] = "" + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.vertices.size();
					smdTable[index][3] = "" + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.vertices.size() + "*";
					smdTable[index][4] = "" + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.vertices.size();
					smdTable[index][5] = "" + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.vertices.size();
					smdTable[index][6] = "" + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.vertices.size();
					smdTable[index][7] = "" + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.vertices.size();
					smdTable[index][8] = "" + e.getValue().legacy.legacy.legacy.legacy.vertices.size();
					smdTable[index][9] = "" + e.getValue().legacy.vertices.size();
					smdTable[index][10] = "" + e.getValue().vertices.size();
					break;
				}
			}
			
			index++;
		}
		
		for (int step = 0; step < 11; step++) {
			String s = "$" + smdTable[0][step] + "$";
			s += " & $" + smdTable[1][step] + "$";
			s += " & $" + smdTable[2][step] + "$";
			s += " & $" + smdTable[3][step] + "$";
			s += " & $" + smdTable[4][step] + "$";
			s += " & $" + smdTable[5][step] + "$";
			s += " & $" + smdTable[6][step] + "$ \\\\";
			System.out.println(s);
		}
		
		for (Map.Entry<JScope, DecaFourStateMachine> e : smPerScope.entrySet()) {
			System.out.println(e.getKey().getName() + ":");
			
			for (Map.Entry<Set<ReprPort>, Integer> x : e.getValue().getFreqPerDivPorts().entrySet()) {
				System.out.println("\t" + Texts.concat(x.getKey(), "+") + " -> " + x.getValue());
			}
		}
		
//		for (Map.Entry<JScope, DecaFourStateMachine> e : smPerScope.entrySet()) {
//			String s = "";
//			s += "7step 8-9-10: " + e.getValue().legacy.legacy.legacy.legacy.vertices.size();
//			s += "6step 7: " + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.vertices.size();
//			s += "5step 6: " + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.vertices.size();
//			s += "4step 4-5: " + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.vertices.size();
//			s += "3step 3: " + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.vertices.size();
//			s += "2step 2: " + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.vertices.size() + "*";
//			s += "1step 1: " + e.getValue().legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.legacy.vertices.size();
//			
//			System.out.println(e.getKey().getName() + " has " + e.getValue().vertices.size() + " vertices and " + e.getValue().transitions.size() + " transitions");
//		}
		
//		System.exit(0);
		
		initialInputs = new PulsePackMap(initialValuePerPort, Dir.IN);
		
		Map<JScope, List<DecaFourVertex>> vtxListPerScope = new HashMap<JScope, List<DecaFourVertex>>();
		
		for (DecaFourStateMachine sm : smPerScope.values()) {
			vtxListPerScope.put(sm.scope, new ArrayList<DecaFourVertex>(sm.vertices));
		}
		
		orderedScopes = new ArrayList<JScope>(smPerScope.keySet());
		initCfg = new DecaFourStateConfig(initialVertices);
		configs = new NormMap<DecaFourStateConfig>();
		vtxPairs = new NormMap<Pair<DecaFourVertex>>();
		vtxs = new NormMap<DecaFourVertex>();
		coverageStats = new DecaFourCoverageStats();
		
		if (name != null && explore) {
			initCfgs();
		} else {
			configs.add(initCfg);
		}
		
		Set<PulsePackMap> outputVals = new HashSet<PulsePackMap>();
		Map<ReprPort, Set<PulsePack>> pulsePacksPerOutputPort = new HashMap<ReprPort, Set<PulsePack>>();
		
		Set<PulsePackMap> envOutputVals = new HashSet<PulsePackMap>();
		Map<ReprPort, Set<PulsePack>> pulsePacksPerEnvOutputPort = new HashMap<ReprPort, Set<PulsePack>>();
		
		for (DecaFourStateConfig cfg : configs) {
			outputVals.add(cfg.getOutputVal());
			envOutputVals.add(cfg.getOutputVal().extractExternalMap());
			
			for (Map.Entry<ReprPort, PulsePack> e : cfg.getOutputVal().getPackPerPort().entrySet()) {
				HashMaps.inject(pulsePacksPerOutputPort, e.getKey(), e.getValue());
				
				if (e.getKey().isPortToEnvironment()) {
					HashMaps.inject(pulsePacksPerEnvOutputPort, e.getKey(), e.getValue());
				}
			}
		}
		
		int multiValueOutputPortCount = 0;
		int constantOutputPortCount = 0;
		
		for (Map.Entry<ReprPort, Set<PulsePack>> e : pulsePacksPerOutputPort.entrySet()) {
			if (e.getValue().size() > 1) {
				multiValueOutputPortCount++;
			} else {
				constantOutputPortCount++;
			}
		}
		
		int multiValueEnvOutputPortCount = 0;
		int constantEnvOutputPortCount = 0;
		
		for (Map.Entry<ReprPort, Set<PulsePack>> e : pulsePacksPerEnvOutputPort.entrySet()) {
			if (e.getValue().size() > 1) {
				multiValueEnvOutputPortCount++;
			} else {
				constantEnvOutputPortCount++;
			}
		}
		
		System.out.println("Environmental input:");
		System.out.println(" #constant-env-input-ports = " + constantInputPortCount); //Repeat
		System.out.println(" #multi-value-env-input-ports = " + multiValueInputPortCount); //Repeat
		System.out.println(" #env-input-labels = " + inputLabelCount); //Repeat
		
		System.out.println("Environmental output:");
		System.out.println(" #constant-env-output-ports = " + constantEnvOutputPortCount);
		System.out.println(" #multi-value-env-output-ports = " + multiValueEnvOutputPortCount);
		System.out.println(" #env-output-labels-potential = " + HashMaps.getCombinationCount(pulsePacksPerEnvOutputPort, false));
		System.out.println(" #env-output-labels-actual = " + envOutputVals.size());
		
		System.out.println("General output:");
		System.out.println(" #constant-output-ports = " + constantOutputPortCount);
		System.out.println(" #multi-value-output-ports = " + multiValueOutputPortCount);
		System.out.println(" #output-labels-potential = " + HashMaps.getCombinationCount(pulsePacksPerOutputPort, false));
		System.out.println(" #output-labels-actual = " + outputVals.size());
		
		for (Pair<DecaFourVertex> vtxPair : vtxPairs) {
			vtxs.add(vtxPair.getElem1());
			vtxs.add(vtxPair.getElem2());
		}
		
		if ("pot-stable".equals("off")) {
			potStableCfgs = extractPotStableCfgs();
		} else {
			potStableCfgs = Collections.emptySet();
		}
		
		System.out.println("#potStableCfgs = " + potStableCfgs.size());
		
		System.out.println("DONE");
		
		System.out.println("#getAllProtoTransitions() = " + getAllProtoTransitions().size());
	}
	
	public void samplePaths() {
		int cfgIndex = 1;
		int detCount = 0;
		int nonDetCount = 0;
		
		for (DecaFourStateConfig cfg : configs) {
			Set<DecaFourInputPath> paths = computePathsFromInitToTarget(cfg);
			
			if (paths.isEmpty()) {
				nonDetCount++;
			} else {
				detCount++;
			}
			
			System.out.println("[" + LocalTime.now() + "] cfg = " + cfgIndex + " / " + configs.size() + "; #det = " + detCount + "; #non-det = " + nonDetCount);
			cfgIndex++;
		}
	}
	
	public void samplePaths2() {
		int cfgIndex = 1;
		
		for (DecaFourStateConfig cfg : configs) {
			System.out.println("[" + LocalTime.now() + "] cfg = " + cfgIndex + " / " + configs.size() + "; computing paths . . .");
			
			Set<DecaFourInputPath> paths = computePathsFromInitToTarget(cfg);
			
			System.out.println("[" + LocalTime.now() + "] cfg = " + cfgIndex + " / " + configs.size() + "; computing potential states . . .");
			
			int nonDetPathCount = 0;
			int detPathCount = 0;
			int idx = 1;
			
			for (DecaFourInputPath path : paths) {
				if (path.isNonDet()) {
					nonDetPathCount++;
				} else {
					detPathCount++;
				}
				
				System.out.println("[" + LocalTime.now() + "] index = " + idx + " / " + paths.size() + "; #det = " + detPathCount + "; #non-det = " + nonDetPathCount);
				idx++;
			}
			
			System.out.println("[" + LocalTime.now() + "] cfg = " + cfgIndex + " / " + configs.size() + "; #paths(" + cfgIndex + ") = " + paths.size() + " (" + nonDetPathCount + " with non-determinism)");
			cfgIndex++;
			
			break;
		}
	}
	
	public Map<Integer, Set<DecaFourInputPath>> computePathsFromInitToTarget2(DecaFourStateConfig targetCfg) {
		Map<Integer, Set<DecaFourInputPath>> result = new HashMap<Integer, Set<DecaFourInputPath>>();
		
		for (DecaFourInputPath path : computePathsFromInitToTarget(targetCfg)) {
			HashMaps.inject(result, path.getNonDetCount(), path);
		}
		
		return result;
	}
	
	public Set<DecaFourInputPath> computePathsFromInitToTarget(DecaFourStateConfig targetCfg) {
		Set<DecaFourInputPath> result = new HashSet<DecaFourInputPath>();
		
		if (targetCfg.level <= 0) {
			result.add(new DecaFourInputPath(targetCfg));
			return result;
		}
		
		for (DecaFourStateConfig predCfg : cfgsPerLevel.get(targetCfg.level - 1)) {
			for (DecaFourStateConfig equivCfg : predCfg.computeSuccs()) {
				if (equivCfg.equals(targetCfg)) {
					PulsePackMap someInput = predCfg.getSomeInputToSucc(equivCfg, true);
					boolean nonDet = false;
					
					if (someInput == null) {
						someInput = predCfg.getSomeInputToSucc(equivCfg, false);
						nonDet = true;
					}
					
					for (DecaFourInputPath predPath : computePathsFromInitToTarget(predCfg)) {
						result.add(new DecaFourInputPath(predPath, someInput, nonDet, equivCfg));
					}
					
					if (result.size() >= 1000) {
						return result;
					}
				}
			}
		}
		
		return result;
	}
	
	public Set<LineageObject> getLineageObjects() {
		Set<LineageObject> result = new HashSet<LineageObject>();
		result.addAll(aggregateLineageObjects());
		return result;
	}
	
	@Override
	public Set<LineageObject> getImpliedElements() {
		Set<LineageObject> result = new HashSet<LineageObject>();
		result.addAll(initProtoTransitions);
		result.addAll(coverageStats.getTransitions4());
		return result;
	}
	
	@Override
	public Set<LineageObject> getInferredElements() {
		return Collections.emptySet();
	}
	
	/**
	 * Union of elements encountered in each of the state machines.
	 */
	public Set<ProtoTransition> getAllProtoTransitions() {
		Set<ProtoTransition> result = new HashSet<ProtoTransition>();
		
		for (DecaFourStateMachine sm : smPerScope.values()) {
			result.addAll(sm.getAllProtoTransitions());
		}
		
		return result;
	}
	
	/**
	 * Union of elements encountered in each of the state machines.
	 */
	public Set<ProtoVertex> getAllProtoVertices() {
		Set<ProtoVertex> result = new HashSet<ProtoVertex>();
		
		for (ProtoTransition t : getAllProtoTransitions()) {
			result.addAll(t.getProtoVertices());
		}
		
		return result;
	}
	
	public Set<DecaFourStateConfig> extractPotStableCfgs() {
		ConcurrentWork<DecaFourStateConfig, ConcurrentWorker> cw = new ConcurrentWork<DecaFourStateConfig, ConcurrentWorker>();
		
		for (int index = 1; index <= 6; index++) {
			ConcurrentWorker w = new ConcurrentWorker();
			w.set("#stable", new HashSet<DecaFourStateConfig>());
			cw.getWorkers().add(w);
		}
		
		cw.apply(configs, (cfg, w) -> {
			if (cfg.couldBeStable()) {
				w.getSet("#stable", DecaFourStateConfig.class).add(cfg);
			}
		});
		
		Set<DecaFourStateConfig> result = new HashSet<DecaFourStateConfig>();
		
		for (ConcurrentWorker w : cw.getWorkers()) {
			result.addAll(w.getSet("#stable", DecaFourStateConfig.class));
		}
		
		return result;
	}
	
//	/**
//	 * Elements encountered in the composition of state machines.
//	 */
//	public Set<ProtoTransition> getAllProtoTransitions2() {
//		return coverageStats.getTransitions();
//	}
//	
//	/**
//	 * Elements encountered in the composition of state machines.
//	 */
//	public Set<ProtoVertex> getAllProtoVertices2() {
//		Set<ProtoVertex> result = new HashSet<ProtoVertex>();
//		
//		for (ProtoTransition t : getAllProtoTransitions2()) {
//			result.addAll(t.getProtoVertices());
//		}
//		
//		return result;
//	}
	
	public void printGraphvizFiles() {
		for (Map.Entry<JScope, DecaFourStateMachine> e : smPerScope.entrySet()) {
			e.getValue().printGraphvizFile();
		}
	}
	
	public double getTotalTrsCount() {
		return totalTrsCount;
	}
	
	public Map<JScope, DecaFourStateMachine> getSmPerScope() {
		return smPerScope;
	}
	
	public void printGraphvizFile(DecaFourStateConfig i, Collection<DecaFourStateConfig> vs) {
		try {
			PrintStream ps = new PrintStream("deca-four-deterministic.gv");
			printGraphvizFile(ps, i, vs);
			ps.flush();
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
	}
	
	public void printGraphvizFile(PrintStream out, DecaFourStateConfig i, Collection<DecaFourStateConfig> vs) {
		Map<DecaFourStateConfig, String> namePerVertex = new HashMap<DecaFourStateConfig, String>();
		Map<DecaFourStateConfig, String> namePerNondetTarget = new HashMap<DecaFourStateConfig, String>();
//		Map<DecaFourStateConfig, String> namePerPeerTarget = new HashMap<DecaFourStateConfig, String>();
		
		for (DecaFourStateConfig v : vs) {
			namePerVertex.put(v, "VTX" + namePerVertex.size());
			namePerNondetTarget.put(v, "X" + namePerNondetTarget.size());
			
//			Set<DecaFourStateConfig> peerTargets = v.getPeerTargets();
//			
//			if (peerTargets.size() > 0) {
//				for (DecaFourStateConfig ptc : peerTargets) {
//					namePerPeerTarget.put(ptc, "PEER" + namePerPeerTarget.size());
//				}
//			}
		}
		
		String initialVertexName = namePerVertex.get(i);
		
		out.println("// " + getClass().getCanonicalName());
		out.println("digraph G {");
		out.println("\tI0 [label=\"(initial)\", shape=circle, fontsize=8];");
		
		for (Map.Entry<DecaFourStateConfig, String> e : namePerVertex.entrySet()) {
			String s = "<" + e.getValue() + " (level " + e.getKey().level + ")>";
			
			for (DecaFourVertex v : e.getKey().getVtxs().values()) {
//				if (e.getKey().sourcePerm != null) {
//					DecaFourTransition t = e.getKey().sourcePerm.get(v.getScope());
//					s += "\\nBEFORE " + t.getSourceVertex().getName();
//					//s += "\\n" + Texts._break(t.getInput().createCombinedGuardStr().replace("\"", "\\\""), "\\n", 100000);
//				}
				
				s += "\\nNOW " + v.getName();
				s += "\\n" + Texts.concat(v.getSysmlClzs(), " + ", (c) -> { return c.getSimpleName(); });
				
				for (Map.Entry<ReprPort, ASALSymbolicValue> entry : v.getOutputVal().extractValuation().entrySet()) {
					if (!entry.getKey().getType().equals(JPulse.class) || entry.getValue().equals(ASALSymbolicValue.TRUE)) {
						s += "\\n" + Texts._break(entry.getKey().getName() + " := " + entry.getValue().toString().replace("\"", "\\\""), "\\n", 80);
					}
				}
			}
			
			out.println("\t" + e.getValue() + " [label=\"" + s + "\", shape=rectangle, fontsize=8];");
		}
		
//		for (Map.Entry<DecaFourStateConfig, String> e : namePerPeerTarget.entrySet()) {
//			String s = "<" + e.getValue() + " (level " + e.getKey().level + ")>";
//			
//			for (DecaFourVertex v : e.getKey().getVtxs().values()) {
//				if (e.getKey().sourcePerm != null) {
//					DecaFourTransition t = e.getKey().sourcePerm.get(v.getScope());
//					s += "\\nBEFORE " + t.getSourceVertex().getName();
//					s += "\\n" + Texts._break(t.getInput().createCombinedGuardStr().replace("\"", "\\\""), "\\n", 100000);
//				}
//				
//				s += "\\nNOW " + v.getName();
//				s += "\\n" + Texts.concat(v.getClzs(), " + ", (c) -> { return c.getSimpleName(); });
//				
//				for (Map.Entry<ReprPort, ASALSymbolicValue> entry : v.getOutputVal().entrySet()) {
//					if (!entry.getKey().getType().equals(JPulse.class) || entry.getValue().equals(ASALSymbolicValue.TRUE)) {
//						s += "\\n" + Texts._break(entry.getKey().getName() + " := " + entry.getValue().toString().replace("\"", "\\\""), "\\n", 80);
//					}
//				}
//			}
//			
//			out.println("\t" + e.getValue() + " [label=\"" + s + "\", shape=rectangle, fontsize=8, style=dashed];");
//		}
		
//		for (DecaFourStateConfig cfg : vs) {
//			Set<DecaFourVertex> tgts = cfg.getNondetTargets();
//			
//			if (tgts.size() > 0) {
//				String s = "<nondet-targets>";
//				
//				for (DecaFourVertex v : tgts) {
//					s += "\\n" + v.getName();
//				}
//				
//				out.println("\t" + namePerNondetTarget.get(cfg) + " [label=\"" + s + "\", shape=rectangle, fontsize=8];");
//				out.println("\t" + namePerVertex.get(cfg) + " -> " + namePerNondetTarget.get(cfg) + " [];");
//			}
//		}
//		
//		int counter = 0;
//		
//		for (DecaFourStateConfig cfg : vs) {
//			
//		}
		
		out.println("\tI0 -> " + initialVertexName + ";");
		out.println("}");
	}
}

