package jlx.behave.proto;

import java.io.*;
import java.util.*;

import jlx.utils.*;

public class LineageObjects {
	private final String name;
	private final Collection<? extends LineageObject> elems;
	private final Map<Class<? extends LineageObject>, Set<LineageObject>> elemsPerClz;
	
	private final Map<String, Set<DecaTwoVertex>> decaTwoVerticesPerComp;
	private final Map<String, Set<DecaTwoTransition>> decaTwoTransitionsPerComp;
	
	public LineageObjects(String name, Collection<? extends LineageObject> elems) {
		this.name = name;
		this.elems = elems;
		
		elemsPerClz = new HashMap<Class<? extends LineageObject>, Set<LineageObject>>();
		
		for (LineageObject elem : elems) {
			HashMaps.inject(elemsPerClz, elem.getClass(), elem);
		}
		
		decaTwoVerticesPerComp = new HashMap<String, Set<DecaTwoVertex>>();
		decaTwoTransitionsPerComp = new HashMap<String, Set<DecaTwoTransition>>();
		
		for (LineageObject elem : elems) {
			if (elem instanceof DecaTwoVertex) {
				DecaTwoVertex v = (DecaTwoVertex)elem;
				HashMaps.inject(decaTwoVerticesPerComp, extractComp(elem), v);
			} else {
				if (elem instanceof DecaTwoTransition) {
					DecaTwoTransition t = (DecaTwoTransition)elem;
					HashMaps.inject(decaTwoTransitionsPerComp, extractComp(elem), t);
				}
			}
		}
	}
	
	private static String extractComp(LineageObject elem) {
		String s = elem.toString();
		int index = s.indexOf("[", 1);
		if (index >= 0) {
			return s.substring(0, index + 1);
		}
		return "??";
	}
	
	public String getName() {
		return name;
	}
	
	public Collection<? extends LineageObject> getElems() {
		return elems;
	}
	
	public Map<Class<? extends LineageObject>, Set<LineageObject>> getElemsPerClz() {
		return elemsPerClz;
	}
	
	public Map<String, Set<DecaTwoVertex>> getDecaTwoVerticesPerComp() {
		return decaTwoVerticesPerComp;
	}
	
	public Map<String, Set<DecaTwoTransition>> getDecaTwoTransitionsPerComp() {
		return decaTwoTransitionsPerComp;
	}
	
	public static Map<Set<LineageObjects>, Map<Class<? extends LineageObject>, Set<LineageObject>>> computeDiffs(LineageObjects... sources) {
		Map<LineageObject, Set<LineageObjects>> sourcesPerElem = new HashMap<LineageObject, Set<LineageObjects>>();
		
		for (LineageObjects source : sources) {
			for (LineageObject elem : source.elems) {
				HashMaps.inject(sourcesPerElem, elem, source);
			}
		}
		
		Map<Set<LineageObjects>, Set<LineageObject>> elemsPerSources = new HashMap<Set<LineageObjects>, Set<LineageObject>>();
		
		for (Map.Entry<LineageObject, Set<LineageObjects>> e : sourcesPerElem.entrySet()) {
			HashMaps.inject(elemsPerSources, e.getValue(), e.getKey());
		}
		
		Map<Set<LineageObjects>, Map<Class<? extends LineageObject>, Set<LineageObject>>> result = new HashMap<Set<LineageObjects>, Map<Class<? extends LineageObject>, Set<LineageObject>>>();
		
		for (Map.Entry<Set<LineageObjects>, Set<LineageObject>> e : elemsPerSources.entrySet()) {
			for (LineageObject elem : e.getValue()) {
				HashMaps.injectInject(result, e.getKey(), elem.getClass(), elem);
			}
		}
		
		return result;
	}
	
	public static Map<Class<? extends LineageObject>, Map<LineageObjects, Set<LineageObject>>> computeFreqs(LineageObjects... sources) {
		Map<LineageObject, Set<LineageObjects>> sourcesPerElem = new HashMap<LineageObject, Set<LineageObjects>>();
		
		for (LineageObjects source : sources) {
			for (LineageObject elem : source.elems) {
				HashMaps.inject(sourcesPerElem, elem, source);
			}
		}
		
		Map<Class<? extends LineageObject>, Map<LineageObjects, Set<LineageObject>>> result = new HashMap<Class<? extends LineageObject>, Map<LineageObjects, Set<LineageObject>>>();
		
		for (Map.Entry<LineageObject, Set<LineageObjects>> e : sourcesPerElem.entrySet()) {
			if (!result.containsKey(e.getKey().getClass())) {
				Map<LineageObjects, Set<LineageObject>> map = new HashMap<LineageObjects, Set<LineageObject>>();
				
				for (LineageObjects source : sources) {
					map.put(source, new HashSet<LineageObject>());
				}
				
				result.put(e.getKey().getClass(), map);
			}
		}
		
		for (Map.Entry<LineageObject, Set<LineageObjects>> e : sourcesPerElem.entrySet()) {
			for (LineageObjects source : e.getValue()) {
				result.get(e.getKey().getClass()).get(source).add(e.getKey());
			}
		}
		
		return result;
	}
	
	public static void printFreqs(PrintStream out, LineageObjects... sources) {
		Map<Class<? extends LineageObject>, Map<LineageObjects, Set<LineageObject>>> freqs = computeFreqs(sources);
		
		{
			String s = "Element type";
			
			for (LineageObjects source : sources) {
				s += "\t" + source.getName();
			}
			
			out.println(s);
		}
		
		for (Map.Entry<Class<? extends LineageObject>, Map<LineageObjects, Set<LineageObject>>> e : freqs.entrySet()) {
			String s = e.getKey().getSimpleName();
			
			for (LineageObjects source : sources) {
				s += "\t" + Integer.toString(e.getValue().get(source).size());
			}
			
			out.println(s);
		}
	}
	
	public static void printFreqs(LineageObjects... sources) {
		printFreqs(System.out, sources);
	}
	
	public static void printFreqs(String filename, LineageObjects... sources) {
		try {
			File targetFile = new File(filename);
			targetFile.getCanonicalFile().getParentFile().mkdirs();
			PrintStream out = new PrintStream(new File(filename));
			printFreqs(out, sources);
			out.close();
		} catch (IOException e) {
			throw new Error(e);
		}
	}
	
	public static void printDiffs(PrintStream out, LineageObjects... sources) {
		Map<Set<LineageObjects>, Map<Class<? extends LineageObject>, Set<LineageObject>>> diffs = computeDiffs(sources);
		
		for (Map.Entry<Set<LineageObjects>, Map<Class<? extends LineageObject>, Set<LineageObject>>> e : diffs.entrySet()) {
			String s;
			
			if (e.getKey().isEmpty()) {
				s = " found nowhere:";
			} else {
				s = " found in " + Texts.concat(e.getKey(), " and ", (x) -> { return x.getName(); }) + ":";
			}
			
			out.println("Objects" + s);
			
			for (Map.Entry<Class<? extends LineageObject>, Set<LineageObject>> e2 : e.getValue().entrySet()) {
				out.println("\t" + e2.getValue().size() + " " + e2.getKey().getSimpleName() + "-objects" + s);
				
				for (LineageObject x : e2.getValue()) {
					out.println("\t\t" + x.toString());
				}
			}
		}
	}
	
	public static void printDiffs(LineageObjects... sources) {
		printDiffs(System.out, sources);
	}
	
	public static void printDiffs(String filename, LineageObjects... sources) {
		try {
			File targetFile = new File(filename);
			targetFile.getCanonicalFile().getParentFile().mkdirs();
			PrintStream out = new PrintStream(new File(filename));
			printDiffs(out, sources);
			out.close();
		} catch (IOException e) {
			throw new Error(e);
		}
	}
	
	private static <T extends LineageObject> Map<Set<LineageObjects>, Set<T>> groupOwners(Map<LineageObjects, Set<T>> elemsPerSource) {
		Map<T, Set<LineageObjects>> ownersPerElem = new HashMap<T, Set<LineageObjects>>();
		
		for (Map.Entry<LineageObjects, Set<T>> e : elemsPerSource.entrySet()) {
			for (T x : e.getValue()) {
				HashMaps.inject(ownersPerElem, x, e.getKey());
			}
		}
		
		Map<Set<LineageObjects>, Set<T>> result = new HashMap<Set<LineageObjects>, Set<T>>();
		
		for (Map.Entry<T, Set<LineageObjects>> e : ownersPerElem.entrySet()) {
			HashMaps.inject(result, e.getValue(), e.getKey());
		}
		
		return result;
	}
	
	public static void printVerticesPerComp(PrintStream out, LineageObjects... sources) {
		Map<String, Map<LineageObjects, Set<DecaTwoVertex>>> verticesPerComp = new HashMap<String, Map<LineageObjects, Set<DecaTwoVertex>>>();
		
		for (LineageObjects source : sources) {
			for (String x : source.getDecaTwoVerticesPerComp().keySet()) {
				HashMaps.injectInjectAll(verticesPerComp, x, source, new HashSet<DecaTwoVertex>());
			}
		}
		
		for (LineageObjects source : sources) {
			for (Map.Entry<String, Set<DecaTwoVertex>> e : source.getDecaTwoVerticesPerComp().entrySet()) {
				HashMaps.injectInjectAll(verticesPerComp, e.getKey(), source, e.getValue());
			}
		}
		
		Map<String, Map<Set<LineageObjects>, Set<DecaTwoVertex>>> sortedVertices = new HashMap<String, Map<Set<LineageObjects>, Set<DecaTwoVertex>>>();
		
		for (Map.Entry<String, Map<LineageObjects, Set<DecaTwoVertex>>> e : verticesPerComp.entrySet()) {
			sortedVertices.put(e.getKey(), groupOwners(e.getValue()));
		}
		
		for (Map.Entry<String, Map<Set<LineageObjects>, Set<DecaTwoVertex>>> e : sortedVertices.entrySet()) {
			out.println(e.getKey() + "-vertices:");
			
			for (Map.Entry<Set<LineageObjects>, Set<DecaTwoVertex>> e2 : e.getValue().entrySet()) {
				String s;
				
				if (e.getKey().isEmpty()) {
					s = " found nowhere";
				} else {
					s = " found in " + Texts.concat(e2.getKey(), " and ", (x) -> { return x.getName(); });
				}
				
				out.println("\t" + e2.getValue().size() + "x " + e.getKey() + "-vertices" + s);
			}
		}
		
		for (Map.Entry<String, Map<Set<LineageObjects>, Set<DecaTwoVertex>>> e : sortedVertices.entrySet()) {
			out.println(e.getKey() + "-vertices:");
			
			for (Map.Entry<Set<LineageObjects>, Set<DecaTwoVertex>> e2 : e.getValue().entrySet()) {
				String s;
				
				if (e.getKey().isEmpty()) {
					s = " found nowhere:";
				} else {
					s = " found in " + Texts.concat(e2.getKey(), " and ", (x) -> { return x.getName(); }) + ":";
				}
				
				out.println("\t" + e.getKey() + "-vertices" + s);
				
				for (DecaTwoVertex v : e2.getValue()) {
					out.println("\t\t" + v.toString());
				}
			}
		}
	}
	
	public static void printVerticesPerComp(LineageObjects... sources) {
		printVerticesPerComp(System.out, sources);
	}
	
	public static void printVerticesPerComp(String filename, LineageObjects... sources) {
		try {
			File targetFile = new File(filename);
			targetFile.getCanonicalFile().getParentFile().mkdirs();
			PrintStream out = new PrintStream(new File(filename));
			printVerticesPerComp(out, sources);
			out.close();
		} catch (IOException e) {
			throw new Error(e);
		}
	}
	
	public static void printTransitionsPerComp(PrintStream out, LineageObjects... sources) {
		Map<String, Map<LineageObjects, Set<DecaTwoTransition>>> transitionsPerComp = new HashMap<String, Map<LineageObjects, Set<DecaTwoTransition>>>();
		
		for (LineageObjects source : sources) {
			for (String x : source.getDecaTwoTransitionsPerComp().keySet()) {
				HashMaps.injectInjectAll(transitionsPerComp, x, source, new HashSet<DecaTwoTransition>());
			}
		}
		
		for (LineageObjects source : sources) {
			for (Map.Entry<String, Set<DecaTwoTransition>> e : source.getDecaTwoTransitionsPerComp().entrySet()) {
				HashMaps.injectInjectAll(transitionsPerComp, e.getKey(), source, e.getValue());
			}
		}
		
		Map<String, Map<Set<LineageObjects>, Set<DecaTwoTransition>>> sortedTransitions = new HashMap<String, Map<Set<LineageObjects>, Set<DecaTwoTransition>>>();
		
		for (Map.Entry<String, Map<LineageObjects, Set<DecaTwoTransition>>> e : transitionsPerComp.entrySet()) {
			sortedTransitions.put(e.getKey(), groupOwners(e.getValue()));
		}
		
		for (Map.Entry<String, Map<Set<LineageObjects>, Set<DecaTwoTransition>>> e : sortedTransitions.entrySet()) {
			out.println(e.getKey() + "-transitions:");
			
			for (Map.Entry<Set<LineageObjects>, Set<DecaTwoTransition>> e2 : e.getValue().entrySet()) {
				String s;
				
				if (e.getKey().isEmpty()) {
					s = " found nowhere";
				} else {
					s = " found in " + Texts.concat(e2.getKey(), " and ", (x) -> { return x.getName(); });
				}
				
				out.println("\t" + e2.getValue().size() + "x " + e.getKey() + "-transitions" + s);
			}
		}
		
		for (Map.Entry<String, Map<Set<LineageObjects>, Set<DecaTwoTransition>>> e : sortedTransitions.entrySet()) {
			out.println(e.getKey() + "-transitions:");
			
			for (Map.Entry<Set<LineageObjects>, Set<DecaTwoTransition>> e2 : e.getValue().entrySet()) {
				String s;
				
				if (e.getKey().isEmpty()) {
					s = " found nowhere:";
				} else {
					s = " found in " + Texts.concat(e2.getKey(), " and ", (x) -> { return x.getName(); }) + ":";
				}
				
				out.println("\t" + e.getKey() + "-transitions" + s);
				
				for (DecaTwoTransition v : e2.getValue()) {
					out.println("\t\t" + v.toString());
				}
			}
		}
	}
	
	public static void printTransitionsPerComp(LineageObjects... sources) {
		printTransitionsPerComp(System.out, sources);
	}
	
	public static void printTransitionsPerComp(String filename, LineageObjects... sources) {
		try {
			File targetFile = new File(filename);
			targetFile.getCanonicalFile().getParentFile().mkdirs();
			PrintStream out = new PrintStream(new File(filename));
			printTransitionsPerComp(out, sources);
			out.close();
		} catch (IOException e) {
			throw new Error(e);
		}
	}
}


