package jlx.behave.proto;

import java.util.*;
import java.util.function.*;

import jlx.behave.stable.*;
import jlx.utils.*;

public class LineageObjectsDiff {
	private final Set<LineageObject> missingElems;
	private final Map<LineageObject, List<DecaStableTransition>> pathToVertexPerMissingElem;
	private final Map<LineageObject, List<DecaStableTransition>> pathToTransitionPerMissingElem;
	
	public LineageObjectsDiff(Collection<? extends LineageObject> subsetElems, Collection<? extends LineageObject> supersetElems, Class<? extends LineageObject> clz, DecaStableStateMachine stableSm) {
		this(subsetElems, supersetElems, (x) -> { return clz.isAssignableFrom(x.getClass()); }, stableSm);
	}
	
	public LineageObjectsDiff(Collection<? extends LineageObject> subsetElems, Collection<? extends LineageObject> supersetElems, Predicate<LineageObject> fct, DecaStableStateMachine stableSm) {
		missingElems = new HashSet<LineageObject>();
		
		for (LineageObject x : supersetElems) {
			if (!subsetElems.contains(x)) {
				if (fct.test(x)) {
					missingElems.add(x);
				}
			}
		}
		
		Set<LineageObject> found = new HashSet<LineageObject>();
		found.addAll(missingElems);
		
		Map<LineageObject, DecaStableVertex> exampleVertexPerMissingElem = new HashMap<LineageObject, DecaStableVertex>();
		
		for (DecaStableVertex v : stableSm.vertices) {
			Set<LineageObject> temp = new HashSet<LineageObject>(v.aggregateLineageObjects());
			temp.retainAll(found);
			
			for (LineageObject tgt : temp) {
				DecaStableVertex w = exampleVertexPerMissingElem.get(tgt);
				
				if (w == null || v.getLevel() < w.getLevel()) {
					exampleVertexPerMissingElem.put(tgt, v);
				}
			}
		}
		
		found.removeAll(exampleVertexPerMissingElem.keySet());
		
		Map<LineageObject, DecaStableTransition> exampleTransitionPerMissingElem = new HashMap<LineageObject, DecaStableTransition>();
		
		for (DecaStableTransition t : stableSm.transitions) {
			Set<LineageObject> temp = new HashSet<LineageObject>(t.aggregateLineageObjects());
			temp.retainAll(found);
			
			for (LineageObject tgt : temp) {
				DecaStableTransition w = exampleTransitionPerMissingElem.get(tgt);
				
				if (w == null || t.getSrc().getLevel() < w.getSrc().getLevel()) {
					exampleTransitionPerMissingElem.put(tgt, t);
				}
			}
		}
		
		pathToVertexPerMissingElem = new HashMap<LineageObject, List<DecaStableTransition>>();
		
		for (Map.Entry<LineageObject, DecaStableVertex> e : exampleVertexPerMissingElem.entrySet()) {
			List<DecaStableTransition> path = new ArrayList<DecaStableTransition>();
			path.addAll(stableSm.getTransitionsToVertex(e.getValue()));
			pathToVertexPerMissingElem.put(e.getKey(), path);
		}
		
		pathToTransitionPerMissingElem = new HashMap<LineageObject, List<DecaStableTransition>>();
		
		for (Map.Entry<LineageObject, DecaStableTransition> e : exampleTransitionPerMissingElem.entrySet()) {
			List<DecaStableTransition> path = new ArrayList<DecaStableTransition>();
			path.addAll(stableSm.getTransitionsToVertex(e.getValue().getSrc()));
			path.add(e.getValue());
			pathToTransitionPerMissingElem.put(e.getKey(), path);
		}
	}
	
	public Set<LineageObject> getMissingElems() {
		return missingElems;
	}
	
	private static void printPath(Lines dest, List<DecaStableTransition> path, String prefix) {
		int index = 1;
		
		for (DecaStableTransition t : path) {
			dest.add(prefix + index + ": " + t.toString().replace("\n", "\n" + prefix + "\t"));
			index++;
		}
	}
	
	private static Map<String, Set<LineageObject>> sortByPrefix(Set<? extends LineageObject> elems) {
		Map<String, Set<LineageObject>> result = new TreeMap<String, Set<LineageObject>>();
		
		for (LineageObject elem : elems) {
			String s = elem.toString();
			int index1 = s.indexOf("[", 1);
			int index2 = s.indexOf("]", 1);
			
			if (index1 >= 0 && index2 >= 0) {
				s = s.substring(0, Math.min(index1, index2) + 1);
			} else {
				if (index1 >= 0) {
					s = s.substring(0, index1 + 1);
				} else {
					if (index2 >= 0) {
						s = s.substring(0, index2 + 1);
					}
				}
			}
			
			HashMaps.inject(result, s, elem);
		}
		
		return result;
	}
	
	private static Map<String, Map<Set<LineageObject>, List<DecaStableTransition>>> sortByPrefix(Map<? extends LineageObject, List<DecaStableTransition>> pathPerElem) {
		Map<String, Map<Set<LineageObject>, List<DecaStableTransition>>> result = new TreeMap<String, Map<Set<LineageObject>, List<DecaStableTransition>>>();
		
		for (Map.Entry<String, Set<LineageObject>> e : sortByPrefix(pathPerElem.keySet()).entrySet()) {
			Map<LineageObject, List<DecaStableTransition>> pathPerElem2 = new HashMap<LineageObject, List<DecaStableTransition>>();
			
			for (LineageObject x : e.getValue()) {
				pathPerElem2.put(x, pathPerElem.get(x));
			}
			
			result.put(e.getKey(), HashMaps.mergeKeysByValues(pathPerElem2));
		}
		
		return result;
	}
	
	public Lines createLines() {
		Lines result = new Lines();
		
		for (LineageObject x : getMissingElems()) {
			if (!pathToVertexPerMissingElem.containsKey(x)) {
				if (!pathToTransitionPerMissingElem.containsKey(x)) {
					result.add("UNREACHABLE-TARGET = " + x.toString());
				}
			}
		}
		
		for (Map.Entry<String, Map<Set<LineageObject>, List<DecaStableTransition>>> e1 : sortByPrefix(pathToVertexPerMissingElem).entrySet()) {
			result.add(e1.getKey() + " (" + e1.getValue().size() + "x, stable):");
			
			for (Map.Entry<Set<LineageObject>, List<DecaStableTransition>> e2 : e1.getValue().entrySet()) {
				for (LineageObject x : e2.getKey()) {
					result.add("\t" + x.toString() + ":");
				}
				
				printPath(result, e2.getValue(), "\t\t");
			}
		}
		
		for (Map.Entry<String, Map<Set<LineageObject>, List<DecaStableTransition>>> e1 : sortByPrefix(pathToTransitionPerMissingElem).entrySet()) {
			result.add(e1.getKey() + " (" + e1.getValue().size() + "x, unstable):");
			
			for (Map.Entry<Set<LineageObject>, List<DecaStableTransition>> e2 : e1.getValue().entrySet()) {
				for (LineageObject x : e2.getKey()) {
					result.add("\t" + x.toString() + ":");
				}
				
				printPath(result, e2.getValue(), "\t\t");
			}
		}
		
		return result;
	}
}

