package jlx.behave.proto.gui;

import java.util.*;
import java.util.function.BiConsumer;

import jlx.asal.j.*;
import jlx.asal.rewrite.ASALSymbolicValue;
import jlx.behave.proto.*;
import jlx.behave.stable.DecaStableOutputEvolution;
import jlx.behave.stable.DecaStableTransition;
import jlx.blocks.ibd1.OutPort;
import jlx.models.UnifyingBlock;
import jlx.models.UnifyingBlock.ReprPort;
import jlx.scenario.*;
import jlx.utils.*;

public class DecaFourScenarioPlayer {
	private String name;
	private final UnifyingBlock unifyingBlock;
	private final Set<DecaFourStateConfig> potVisitedCfgs;
	private final Set<DecaFourStateConfig> lookedOutForCfgs;
	private BiConsumer<DecaFourStateConfig, DecaFourScenarioLog> cfgConsumer;
	private final DecaFourCoverageStats coverageStats;
	private final List<DecaStableTransition> stableTrs;
	private final Set<DecaFourCfgPair> cfgPairs;
	private final Map<DecaFourCfgPair, Set<Integer>> steps;
	private int noStableTrCount;
	private int stepCount;
	private final Map<DecaFourStateConfig, Set<Step.MetadataStep>> stepsPerCfg;
	
	public DecaFourScenarioPlayer(UnifyingBlock unifyingBlock) {
		this.unifyingBlock = unifyingBlock;
		
		name = "unnamed-scenario-player";
		potVisitedCfgs = new HashSet<DecaFourStateConfig>();
		lookedOutForCfgs = new HashSet<DecaFourStateConfig>();
		cfgConsumer = null;
		coverageStats = new DecaFourCoverageStats();
		stableTrs = new ArrayList<DecaStableTransition>();
		cfgPairs = new HashSet<DecaFourCfgPair>();
		steps = new HashMap<DecaFourCfgPair, Set<Integer>>();
		noStableTrCount = 0;
		stepCount = 0;
		stepsPerCfg = new HashMap<DecaFourStateConfig, Set<Step.MetadataStep>>();
	}
	
	public String getName() {
		return name;
	}
	
	public void setName(String name) {
		this.name = name;
	}
	
	public UnifyingBlock getUnifyingBlock() {
		return unifyingBlock;
	}
	
	/**
	 * Visited configurations of the composed FSM.
	 * 
	 * "Potential" because the same input + output evolutions can, in principle, be achieved via multiple routes.
	 * We assume that ALL routes (and all configurations along the routes) are visited.
	 */
	public Set<DecaFourStateConfig> getPotVisitedCfgs() {
		return potVisitedCfgs;
	}
	
	/**
	 * Transitions in the SIC-DFSM that were visited.
	 * The SIC-DFSM must have been computed in advance.
	 */
	public List<DecaStableTransition> getStableTrs() {
		return stableTrs;
	}
	
	/**
	 * Number of times that an observed input + output evolutions did not match a transition in the SIC-DFSM.
	 * Increased continuously when the SIC-DFSM was not computed in advance.
	 */
	public int getNoStableTrCount() {
		return noStableTrCount;
	}
	
	public int getStepCount() {
		return stepCount;
	}
	
	/**
	 * Contains, for a given pair of configurations (A, B), the indices of the test steps during which they were observed.
	 * 
	 * NOTE: The test step index is not reset in between tests, should this happen...?
	 */
	public Map<DecaFourCfgPair, Set<Integer>> getSteps() {
		return steps;
	}
	
	public Map<DecaFourStateConfig, Set<Step.MetadataStep>> getStepsPerCfg() {
		return stepsPerCfg;
	}
	
	public Set<LineageObject> getLineageObjects() {
		Set<LineageObject> result = new HashSet<LineageObject>();
		
		//Initial transitions have been pre-processed.
		//Since we (assume that we) cover them, we add them explicitly:
		for (LineageObject t : unifyingBlock.sms4.initProtoTransitions) {
			result.addAll(t.aggregateLineageObjects());
			result.add(t);
		}
		
		for (LineageObject t : coverageStats.getTransitions4()) {
			result.addAll(t.aggregateLineageObjects());
			result.add(t);
		}
		
		return result;
	}
	
	/**
	 * Contains a pair of configurations (A, B) if we observed an input + output evolutions that led from A to B.
	 */
	public Set<DecaFourCfgPair> getCfgPairs() {
		 return cfgPairs;
	}
	
	public DecaFourScenarioLog play(Scenario scenario, String failVerdictLogFilename) {
		if (scenario.getAltSteps() != null) {
			play(scenario, null, false);
			return play(scenario, failVerdictLogFilename, true);
		}
		
		return play(scenario, failVerdictLogFilename, false);
	}
	
	private DecaFourScenarioLog play(Scenario scenario, String failVerdictLogFilename, boolean useAltSteps) {
		DecaFourScenarioLog result = play(scenario, useAltSteps);
		
		if (result.getMismatches().size() > 0) {
			if (failVerdictLogFilename != null) {
				result.saveToFile(failVerdictLogFilename);
			} else {
				for (String line : result.getLines()) {
					System.out.println(line);
				}
				
				CLI.waitForEnter();
			}
		}
		
		return result;
	}
	
//	private Map<ReprPort, ASALSymbolicValue> extractInitialInputVal() {
//		Map<ReprPort, ASALSymbolicValue> valuePerPort = new HashMap<ReprPort, ASALSymbolicValue>();
//		
//		for (Map.Entry<ReprPort, PulsePack> e : unifyingBlock.sms4.initialInputs.getPackPerPort().entrySet()) {
//			if (e.getKey().isPortToEnvironment()) {
//				valuePerPort.put(e.getKey(), e.getValue());
//			}
//		}
//		
//		return valuePerPort;
//	}
	
	private DecaFourScenarioLog play(Scenario scenario, boolean useAltSteps) {
		DecaFourScenarioLog result = internalPlay(scenario, useAltSteps);
		
		for (Set<DecaFourStateConfig> cfgs : result.getPotCfgsSeq()) {
			if (potVisitedCfgs.addAll(cfgs)) {
				throw new Error("?!");
			}
		}
		
		coverageStats.addAll(result.getCoverageStats());
		stableTrs.addAll(result.getStableTrs());
		noStableTrCount += result.getNoStableTrCount();
		stepCount += result.getInputValSeq().size();
		cfgPairs.addAll(result.getCfgPairs());
		
		List<DecaFourCfgPair> cfgPairSeq = new ArrayList<DecaFourCfgPair>();
		result.populateCfgPairs(cfgPairSeq);
		
		for (int index = 0; index < cfgPairSeq.size(); index++) {
			DecaFourCfgPair cfgPair = cfgPairSeq.get(index);
			Set<Integer> dest = steps.get(cfgPair);
			
			if (dest == null) {
				dest = new TreeSet<Integer>();
				steps.put(cfgPair, dest);
			}
			
			dest.add(index + 1);
		}
		
		return result;
	}
	
	private final static String INDENT = "  ";
	
	private DecaFourScenarioLog internalPlay(Scenario scenario, boolean useAltSteps) {
		potVisitedCfgs.add(unifyingBlock.sms4.initCfg);
		HashMaps.injectAll(stepsPerCfg, unifyingBlock.sms4.initCfg, Collections.emptySet());
		
		DecaFourScenarioLog result = new DecaFourScenarioLog(scenario, useAltSteps, unifyingBlock.sms4.initCfg);
		result.getLines().add("========== " + scenario.getClass().getCanonicalName() + " ==========");
		result.getLines().add("Starting in state");
		
		for (Map.Entry<JScope, DecaFourVertex> e : unifyingBlock.sms4.initCfg.getVtxs().entrySet()) {
			result.getLines().add(INDENT + e.getValue().getName());
		}
		
		result.getLines().add("Initial input values:");
		
		for (Map.Entry<ReprPort, PulsePack> e : unifyingBlock.sms4.initialInputs.getPackPerPort().entrySet()) {
			result.getLines().add(INDENT + e.getValue());
		}
		
		result.getLines().add("Initial output values:");
		
		for (Map.Entry<ReprPort, PulsePack> e : unifyingBlock.sms4.initCfg.getOutputVal().getPackPerPort().entrySet()) {
			result.getLines().add(INDENT + e.getValue());
		}
		
		Map<ReprPort, ASALSymbolicValue> currInputs = unifyingBlock.sms4.initialInputs.extractExternalMap().extractValuation();
		Set<DecaStableOutputEvolution> expectedOutputs = new HashSet<DecaStableOutputEvolution>();
		Set<String> expectedStateChanges = new TreeSet<String>();
		List<Step.MetadataStep> metadataSteps = new ArrayList<Step.MetadataStep>();
		Step[] steps = useAltSteps ? scenario.getAltSteps() : scenario.getSteps();
		
		for (int index = 0; index < steps.length; index++) {
			Step step = steps[index];
			
			if (result.getLastPotCfgs().isEmpty()) {
				throw new Error("Should not happen!");
			}
			
			try {
			if (step instanceof Step.InputStep) {
				Step.InputStep i = (Step.InputStep)step;
				ReprPort rp = unifyingBlock.reprPortPerPrimPort.get(i.getPort());
				
				if (rp == null) {
					throw new Error("Could not find input port" + i.getPort().getFileLocation());
				}
				
				if (!rp.isPortToEnvironment()) {
					throw new Error("Input port should be an environmental input port:" + i.getPort().getFileLocation());
				}
				
				ASALSymbolicValue v = ASALSymbolicValue.from(i.getNewValue());
				
				//Do not set duration ports of the specification FSM, since they are not part of input valuation!! 
				if (rp.isDurationPort()) {
					result.getLines().add("( Ignoring assignment of " + v.toString() + " to duration port " + rp.getReprOwner().getName() + "::" + rp.getName() + " )");
					continue;
				}
				
				result.getLines().add("Setting " + rp.getReprOwner().getName() + "::" + rp.getName() + " to " + v.toString());
				currInputs.put(rp, v);
				continue;
			}
			
			if (step instanceof Step.TimeoutStep) {
				Step.TimeoutStep t = (Step.TimeoutStep)step;
				ReprPort rp = unifyingBlock.reprPortPerPrimPort.get(t.getDurationPort());
				
				if (rp == null) {
					throw new Error("Could not find duration port" + t.getDurationPort().getFileLocation());
				}
				
				ReprPort rp2 = getTimeoutPort(rp);
				
				if (rp2 == null) {
					throw new Error("Could not find timeout port that corresponds with duration port" + rp.getFileLocation());
				}
				
				result.getLines().add("Triggering timeout " + rp2.getReprOwner().getName() + "::" + rp2.getName());
				currInputs.put(rp2, ASALSymbolicValue.TRUE);
				continue;
			}
			
			if (step instanceof Step.OutputStep) {
				Step.OutputStep o = (Step.OutputStep)step;
				List<PulsePack> outputSeq = new ArrayList<PulsePack>();
				
				for (Map<OutPort<?>, JType> e : o.getOutputSeq()) {
					Map<ReprPort, ASALSymbolicValue> m = new HashMap<ReprPort, ASALSymbolicValue>();
					ReprPort mainPort = null;
					
					for (Map.Entry<OutPort<?>, JType> e2 : e.entrySet()) {
						ReprPort rp = unifyingBlock.reprPortPerPrimPort.get(e2.getKey());
						
						if (rp == null) {
							throw new Error("Could not find output port" + e2.getKey().getFileLocation());
						}
						
						m.put(rp, ASALSymbolicValue.from(e2.getValue()));
						
						if (rp.getPulsePort() == null) {
							if (mainPort != null) {
								throw new Error("Pulse pack has too many main ports" + e2.getKey().getFileLocation());
							}
							
							mainPort = rp;
						}
					}
					
					if (outputSeq.isEmpty() || !outputSeq.get(outputSeq.size() - 1).getValuePerPort().equals(m)) {
						if (mainPort == null) {
							throw new Error("Pulse pack has no main ports");
						}
						
						outputSeq.add(new PulsePack(mainPort, m));
					}
				}
				
				DecaStableOutputEvolution evo = new DecaStableOutputEvolution(outputSeq);
				result.getLines().add(INDENT + "Expecting " + evo);
				expectedOutputs.add(evo);
				continue;
			}
			
			if (step instanceof Step.SingleStep) {
				result.getLines().add("Observing 1 step");
				result.getLines().addAll(expectedStateChanges);
				
				PulsePackMap inputVal = PulsePackMap.from(currInputs, Dir.IN);
				
//				if (unifyingBlock.stableSm.transitions.size() > 0) {
//					boolean noStableTr = true;
//					
//					for (DecaStableTransition t : unifyingBlock.stableSm.transitions) {
//						if (t.getSicInputs1() != null) {
//							if (result.getLastPotCfgs().contains(t.getSrc().getCfg())) {
//								if (t.getSicInputs1().equals(inputVal)) {
//									result.getStableTrs().add(t);
//									noStableTr = false;
//								}
//							}
//						}
//					}
//					
//					if (noStableTr) {
//						result.incNoStableTrCount();
//					}
//				}
				
				Set<DecaFourStateConfig> newPotCfgs = new HashSet<DecaFourStateConfig>();
				Set<DecaStableOutputEvolution> evos = new HashSet<DecaStableOutputEvolution>();
				Set<DecaFourScenarioMismatch> leastWrongMismatches = null;
				
				for (DecaFourStateConfig potCfg : result.getLastPotCfgs()) {
					Set<DecaFourStateConfig> succs = potCfg.computeSuccsViaInputVal(inputVal).keySet();
					result.getCoverageStats().addAll(potCfg.getCoverageStatsToSuccsViaInputVal(succs, inputVal));
					
					for (DecaFourStateConfig succ : succs) {
						Set<DecaStableOutputEvolution> evos__ = DecaStableOutputEvolution.getOutputEvolutions(potCfg, succ);
						Set<DecaFourScenarioMismatch> mismatches = getMismatches(evos__, expectedOutputs);
						
						if (mismatches.isEmpty()) {
							newPotCfgs.add(succ);
							evos.addAll(evos__);
						} else {
							if (leastWrongMismatches == null || mismatches.size() < leastWrongMismatches.size()) {
								leastWrongMismatches = mismatches;
							}
						}
					}
				}
				
				if (newPotCfgs.isEmpty()) {
					addMismatchesToLog(leastWrongMismatches, result);
					result.getMismatches().addAll(leastWrongMismatches);
					result.setStepFileLocation(step.getFileLocation());
					result.getLines().add("Fail verdict");
					return result;
				}
				
				for (DecaStableOutputEvolution evo : evos) {
					if (evo.getEvolution().size() > 1) {
						result.getLines().add(INDENT + "Observed " + evo);
					}
				}
				
				addNewPotCfgsToLog(newPotCfgs, metadataSteps, result);
				potVisitedCfgs.addAll(newPotCfgs);
				
				for (DecaFourStateConfig potCfg : newPotCfgs) {
					result.getLines().add(INDENT + "Arrived at " + potCfg);
					HashMaps.injectAll(stepsPerCfg, potCfg, metadataSteps);
				}
				
				result.getPotCfgsSeq().add(newPotCfgs);
				result.getInputValSeq().add(inputVal);
				currInputs = inputVal.deactivate().extractValuation();
				expectedOutputs.clear();
				expectedStateChanges.clear();
				metadataSteps.clear();
				continue;
			}
			
			if (step instanceof Step.StabilizeStep) {
				result.getLines().addAll(expectedStateChanges);
				result.getLines().add("Observing until stable");
				
				PulsePackMap inputVal = PulsePackMap.from(currInputs, Dir.IN);
				
//				if (result.getInputValSeq().size() > 0 && unifyingBlock.stableSm.transitions.size() > 0) {
//					boolean noStableTr = true;
//					
//					for (DecaStableTransition t : unifyingBlock.stableSm.transitions) {
//						if (t.getSicInputs1() != null) {
//							if (result.getLastPotCfgs().contains(t.getSrc().getCfg())) {
//								if (t.getSicInputs1().equals(inputVal)) {
//									result.getStableTrs().add(t);
//									noStableTr = false;
//								}
//							}
//						}
//					}
//					
//					if (noStableTr) {
//						result.incNoStableTrCount();
//						
//						if (result.getInputValSeq().size() > 0) {
//							System.out.println("I1: " + result.getInputValSeq().get(result.getInputValSeq().size() - 1).toString());
//							result.getLines().add("I1: " + result.getInputValSeq().get(result.getInputValSeq().size() - 1).toString());
//						} else {
//							System.out.println("I1: NULL");
//							result.getLines().add("I1: NULL");
//						}
//						
//						System.out.println("I2: " + inputVal.toString());
//						result.getLines().add("I2: " + inputVal.toString());
//						CLI.waitForEnter();
//					}
//				}
				
				Set<DecaFourStateConfig> potVisited = new HashSet<DecaFourStateConfig>();
				Set<DecaFourStateConfig> newPotCfgs = new HashSet<DecaFourStateConfig>();
				Set<DecaStableOutputEvolution> evos = new HashSet<DecaStableOutputEvolution>();
				Set<DecaFourScenarioMismatch> leastWrongMismatches = null;
				
				for (DecaFourStateConfig potCfg : result.getLastPotCfgs()) {
					Set<DecaFourStateConfig> succs = potCfg.computeSuccsViaInputVal(inputVal).keySet();
					
					if (succs.size() > 1) {
						result.incNonDetTrCount();
					}
					
//					if (succs.size() > 0) {
						for (DecaFourStateConfigPath p : DecaFourStateConfig.followInputs2(succs, inputVal.deactivate(), inputVal)) {
							List<DecaFourStateConfig> seq = new ArrayList<DecaFourStateConfig>(p.getCfgs());
							seq.add(0, potCfg);
							
							Set<DecaStableOutputEvolution> evos__ = DecaStableOutputEvolution.getOutputEvolutions(seq);
							Set<DecaFourScenarioMismatch> mismatches = getMismatches(evos__, expectedOutputs);
							
							if (mismatches.isEmpty()) {
								result.getCoverageStats().addAll(p.getCoverageStats());
								result.getCoverageStats().addAll(potCfg.getCoverageStatsToSuccsViaInputVal(succs, inputVal));
								
								potVisited.addAll(seq);
								newPotCfgs.add(seq.get(seq.size() - 1));
								evos.addAll(evos__);
							} else {
								if (leastWrongMismatches == null || mismatches.size() < leastWrongMismatches.size()) {
									leastWrongMismatches = mismatches;
								}
							}
						}
						
//						List<DecaFourStateConfig> seq = DecaFourStateConfig.followInputs(succs, inputVal.deactivate(), inputVal).getCfgs();
//						
//						if (seq == null) {
//							seq = new ArrayList<DecaFourStateConfig>();
//							seq.add(potCfg);
//						}
//						
//						seq.add(0, potCfg);
//						DecaFourStateConfig succ = seq.get(seq.size() - 1);
//						Set<DecaStableOutputEvolution> evos__ = DecaStableOutputEvolution.getOutputEvolutions(seq);
//						Set<DecaFourScenarioMismatch> mismatches = getMismatches(evos__, expectedOutputs);
//						
//						if (mismatches.isEmpty()) {
//							potVisited.addAll(seq);
//							result.getProtoTrs().addAll(seq.get)
//							newPotCfgs.add(succ);
//							evos.addAll(evos__);
//						} else {
//							if (leastWrongMismatches == null || mismatches.size() < leastWrongMismatches.size()) {
//								leastWrongMismatches = mismatches;
//							}
//						}
//					}
				}
				
				if (newPotCfgs.isEmpty()) {
					addMismatchesToLog(leastWrongMismatches, result);
					result.getPotCfgsSeq().add(newPotCfgs);
					
					result.getMismatches().addAll(leastWrongMismatches);
					result.setStepFileLocation(step.getFileLocation());
					result.getLines().add("Fail verdict");
					return result;
				}
				
				for (DecaStableOutputEvolution evo : evos) {
					if (evo.getEvolution().size() > 1) {
						result.getLines().add(INDENT + "Observed " + evo);
					}
				}
				
				addNewPotCfgsToLog(newPotCfgs, metadataSteps, result);
				potVisitedCfgs.addAll(potVisited);
				
				if (cfgConsumer != null) {
					for (DecaFourStateConfig potCfg : potVisited) {
						cfgConsumer.accept(potCfg, result);
					}
				}
				
				for (DecaFourStateConfig potCfg : potVisited) {
					result.getLines().add(INDENT + "Arrived at " + potCfg);
					HashMaps.injectAll(stepsPerCfg, potCfg, metadataSteps);
				}
				
				for (Step.MetadataStep metadataStep : metadataSteps) {
					result.getLines().add(metadataStep.oneLine(unifyingBlock));
				}
				
				result.getPotCfgsSeq().add(newPotCfgs);
				result.getInputValSeq().add(inputVal);
				currInputs = inputVal.deactivate().extractValuation();
				expectedOutputs.clear();
				expectedStateChanges.clear();
				metadataSteps.clear();
				continue;
			}
			
			if (step instanceof Step.StateChange) {
				Step.StateChange s = (Step.StateChange)step;
				
				String name = s.getBlockInstance().getName();
//				String name = unifyingBlock.reprBlockPerNarrowBlock.get(s.getBlockInstance()).getName(); //Does not work for DB test suite??!
				
				if (s.getStateSeq().size() == 1) {
					expectedStateChanges.add(INDENT + name + " should stay in " + s.getStateSeq().get(0));
				} else {
					expectedStateChanges.add(INDENT + name + " should move from " + Texts.concat(s.getStateSeq(), " to "));
				}
				
				continue;
			}
			
			if (step instanceof Step.MetadataStep) {
				Step.MetadataStep s = (Step.MetadataStep)step;
				metadataSteps.add(s);
				continue;
			}
			} catch (Throwable e) {
				String s = "";
				
				for (int idx = 0; idx <= index; idx++) {
					s += "\n\t\t" + steps[idx].oneLine(unifyingBlock);
				}
				
				throw new Error("Unanticipated failure after:" + s, e);
			}
			
			throw new Error("Should not happen!");
		}
		
		result.getLines().add("Pass verdict");
//		CLI.waitForEnter();
		return result;
	}
	
	private static Set<DecaFourScenarioMismatch> getMismatches(Set<DecaStableOutputEvolution> found, Set<DecaStableOutputEvolution> expected) {
		Set<DecaFourScenarioMismatch> result = new HashSet<DecaFourScenarioMismatch>();
		
		for (DecaStableOutputEvolution evo : found) {
			DecaFourScenarioMismatch mismatch = getMismatch(evo, expected);
			
			if (mismatch != null) {
				result.add(mismatch);
			}
		}
		
		return result;
	}
	
	private static DecaFourScenarioMismatch getMismatch(DecaStableOutputEvolution found, Set<DecaStableOutputEvolution> expected) {
		//We found an output that did NOT evolve:
		if (found.getEvolution().size() == 1) {
			//Therefore, it should not be one of the expected output evolutions:
			for (DecaStableOutputEvolution evo : expected) {
				if (evo.getEvolution().get(0).equals(found.getEvolution().get(0))) {
					if (evo.getEvolution().size() > 1) { //Ignore "non-evolving evolutions"
						return new DecaFourScenarioMismatch("(!) Incomplete output evolution", found, evo);
					}
				}
			}
			
			return null;
		}
		
		for (DecaStableOutputEvolution evo : expected) {
			int sharedPrefixLength = DecaStableOutputEvolution.getSharedPrefixLength(evo, found);
			
			if (sharedPrefixLength > 0) {
				if (sharedPrefixLength == evo.getEvolution().size() && sharedPrefixLength == found.getEvolution().size()) {
					return null;
				}
				
				if (sharedPrefixLength == found.getEvolution().size()) {
					return new DecaFourScenarioMismatch("(!) Incomplete output evolution", found, evo);
				}
				
				if (sharedPrefixLength == evo.getEvolution().size()) {
					return new DecaFourScenarioMismatch("(!) Unexpected output evolution", found, evo);
				}
				
				return new DecaFourScenarioMismatch("(!) Different output evolution", found, evo);
			}
		}
		
		DecaStableOutputEvolution evo = new DecaStableOutputEvolution(Collections.singletonList(found.getEvolution().get(0)));
		return new DecaFourScenarioMismatch("(!) Unexpected output evolution", found, evo);
	}
	
	private ReprPort getTimeoutPort(ReprPort durationPort) {
		DecaFourStateMachine sm = unifyingBlock.sms4.getSmPerScope().get(durationPort.getReprOwner());
		return (ReprPort)sm.timeoutPortPerDurationPort.get(durationPort);
	}
	
	private static void addMismatchesToLog(Set<DecaFourScenarioMismatch> mismatches, DecaFourScenarioLog log) {
		for (DecaFourScenarioMismatch mismatch : mismatches) {
			log.getLines().add(mismatch.getHeader() + ":");
			log.getLines().add(INDENT + "Expected: " + mismatch.getExpected());
			log.getLines().add(INDENT + "Observed: " + mismatch.getFound());
		}
	}
	
	public Set<DecaFourStateConfig> getLookedOutForCfgs() {
		return lookedOutForCfgs;
	}
	
	public void setCfgConsumer(BiConsumer<DecaFourStateConfig, DecaFourScenarioLog> cfgConsumer) {
		this.cfgConsumer = cfgConsumer;
	}
	
	private void addNewPotCfgsToLog(Set<DecaFourStateConfig> newPotCfgs, List<Step.MetadataStep> metadataSteps, DecaFourScenarioLog log) {
		Iterator<DecaFourStateConfig> q = newPotCfgs.iterator();
		addNewPotCfgToLog("Moving to state", "Moving to flagged state", q.next(), metadataSteps, log);
		
		while (q.hasNext()) {
			addNewPotCfgToLog("or to state", "or moving to flagged state", q.next(), metadataSteps, log);
		}
	}
	
	private void addNewPotCfgToLog(String msg1, String msg2, DecaFourStateConfig newPotCfg, List<Step.MetadataStep> metadataSteps, DecaFourScenarioLog log) {
		if (lookedOutForCfgs.contains(newPotCfg)) {
			log.getFoundCfgs().add(newPotCfg);
			log.getLines().add(msg2);
		} else {
			log.getLines().add(msg1);
		}
		
		for (Map.Entry<JScope, DecaFourVertex> e : newPotCfg.getVtxs().entrySet()) {
			log.getLines().add(INDENT + e.getValue().getName());
		}
		
		if (cfgConsumer != null) {
			cfgConsumer.accept(newPotCfg, log);
		}
	}
}

