package jlx.behave.stable.files;

import java.io.*;
import java.time.LocalTime;
import java.util.*;

import jlx.utils.*;

public class DecaStableFileWriter extends DecaStableFilePrinter {
	public void saveToFile(String dirName, String fileName) {
		try {
			File dir = new File(dirName);
			System.out.println("[" + LocalTime.now() + "] Saving to " + dir.getAbsolutePath() + "...");
			
			if (dir.mkdir()) {
				System.out.println("Created directory \"" + dir.getAbsolutePath() + "\" for storing " + getClass().getCanonicalName() + "!");
			}
			
			DataOutputStream out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(dirName + "/" + fileName)));
			writeToStream(out);
			out.flush();
			out.close();
			
			System.out.println("[" + LocalTime.now() + "] Saving to " + dir.getAbsolutePath() + "... done!");
		} catch (IOException e) {
			throw new Error(e);
		}
	}
	
	public List<DecaStableFileContent> getOrderedContents() {
		List<DecaStableFileContent> result = new ArrayList<DecaStableFileContent>();
		
		//Resolution of objects depends heavily on this ordering:
		result.add(DecaStableFileContent.SCOPES);
		result.add(DecaStableFileContent.INPUT_PORTS);
		result.add(DecaStableFileContent.OUTPUT_PORTS);
		result.add(DecaStableFileContent.PROTO);
		result.add(DecaStableFileContent.VERTICES);
		result.add(DecaStableFileContent.INPUT_CHANGES);
		result.add(DecaStableFileContent.OUTPUT_EVOS);
		result.add(DecaStableFileContent.TRANSITIONS);
		
		if (getCharSetPerVertex().size() > 0) {
			result.add(DecaStableFileContent.CHARSETS);
		}
		
		if (getTestSuites().size() > 0) {
			result.add(DecaStableFileContent.TEST_SUITES);
		}
		
		return result;
	}
	
	private void writeToStream(DataOutputStream out) throws IOException {
		TextOptions.select(TextOptions.FULL);
		
		List<DecaStableFileContent> orderedContents = getOrderedContents();
		out.writeInt(orderedContents.size());
		
		for (DecaStableFileContent c : orderedContents) {
			out.writeUTF(c.getId());
		}
		
		for (DecaStableFileContent c : orderedContents) {
			switch (c) {
				case CHARSETS:
					if (getCharSetPerVertex().size() != getVertices().size()) {
						throw new Error("Should not happen!");
					}
					
					for (Vertex vertex : getVertices().values()) {
						CharSet charSet = getCharSetPerVertex().get(vertex);
						writeCharSet(out, charSet);
					}
					
					break;
				case INPUT_CHANGES:
					out.writeInt(getInputChanges().size());
					
					for (Map.Entry<Integer, InputChanges> e : getInputChanges().entrySet()) {
						writeInputChanges(out, e.getValue());
					}
					
					break;
				case INPUT_PORTS:
					out.writeInt(getInputPorts().size());
					
					for (Map.Entry<Integer, Port> e : getInputPorts().entrySet()) {
						writePort(out, e.getValue());
					}
					
					break;
				case OUTPUT_EVOS:
					out.writeInt(getOutputEvolutions().size());
					
					for (Map.Entry<Integer, OutputEvolution> e : getOutputEvolutions().entrySet()) {
						writeOutputEvolution(out, e.getValue());
					}
					
					break;
				case OUTPUT_PORTS:
					out.writeInt(getOutputPorts().size());
					
					for (Map.Entry<Integer, Port> e : getOutputPorts().entrySet()) {
						writePort(out, e.getValue());
					}
					
					break;
				case SCOPES:
					out.writeInt(getScopes().size());
					
					for (Map.Entry<Integer, Scope> e : getScopes().entrySet()) {
						writeScope(out, e.getValue());
					}
					
					break;
				case TEST_SUITES:
					out.writeInt(getTestSuites().size());
					
					for (TestSuite testSuite : getTestSuites()) {
						writeTestSuite(out, testSuite);
					}
					
					break;
				case VERTICES:
					out.writeInt(getVertices().size());
					
					for (Map.Entry<Integer, Vertex> e : getVertices().entrySet()) {
						writeVertex(out, e.getValue());
					}
					
					break;
				case TRANSITIONS:
					writeTransition(out, getInitialTransition());
					out.writeInt(getInitialTransition().getTgt().getId());
					out.writeInt(getVertices().size());
					
					for (Map.Entry<Integer, Vertex> e1 : getVertices().entrySet()) {
						SortedMap<Integer, List<Transition>> transitionsPerTgtId = e1.getValue().getTransitionsPerTgtId();
						
						out.writeInt(e1.getKey());
						out.writeInt(transitionsPerTgtId.size());
						
						for (Map.Entry<Integer, List<Transition>> e : transitionsPerTgtId.entrySet()) {
							out.writeInt(e.getKey());
							out.writeInt(e.getValue().size());
							
							for (Transition transition : e.getValue()) {
								writeTransition(out, transition);
							}
						}
					}
					
					break;
				case PROTO:
					out.writeInt(getProto().size());
					
					for (Map.Entry<Integer, Proto> e : getProto().entrySet()) {
						writeProto(out, e.getValue());
					}
					
					out.writeInt(getProtoSets().size());
					
					for (Map.Entry<Integer, ProtoSet> e : getProtoSets().entrySet()) {
						writeProtoSet(out, e.getValue());
					}
					
					break;
				default:
					throw new Error("Should not happen!");
			}
		}
	}
	
	private void writeScope(DataOutputStream out, Scope scope) throws IOException {
		out.writeUTF(scope.getName());
	}
	
	private void writePort(DataOutputStream out, Port port) throws IOException {
		out.writeUTF(port.getName());
		out.writeInt(port.getOwner().getId());
		out.writeUTF(port.getTypeName());
		out.writeInt(port.getPortIndex());
		out.writeUTF(port.getAdapterLabel());
		out.writeBoolean(port.isPortToEnvironment());
		out.writeBoolean(port.isTimeoutPort());
		out.writeInt(port.getExecutionTime());
	}
	
	private void writeVertex(DataOutputStream out, Vertex vertex) throws IOException {
		SortedMap<Integer, String> statePerScope = new TreeMap<Integer, String>();
		
		for (Map.Entry<Scope, String> e : vertex.getStatePerScope().entrySet()) {
			statePerScope.put(e.getKey().getId(), e.getValue());
		}
		
		for (Map.Entry<Integer, String> e : statePerScope.entrySet()) {
			out.writeUTF(e.getValue());
		}
		
		SortedMap<Integer, String> clzsPerScope = new TreeMap<Integer, String>();
		
		for (Map.Entry<Scope, String> e : vertex.getClzsPerScope().entrySet()) {
			clzsPerScope.put(e.getKey().getId(), e.getValue());
		}
		
		for (Map.Entry<Integer, String> e : clzsPerScope.entrySet()) {
			out.writeUTF(e.getValue());
		}
		
		out.writeInt(vertex.getProtoSet().getId());
	}
	
	private void writeInputChanges(DataOutputStream out, InputChanges inputChanges) throws IOException {
		out.writeInt(inputChanges.getNewValuePerPort().size());
		
		for (Map.Entry<Port, String> e2 : inputChanges.getNewValuePerPort().entrySet()) {
			out.writeInt(e2.getKey().getId());
			out.writeUTF(e2.getValue());
		}
		
		if (inputChanges.getDurationPort() != null) {
			out.writeBoolean(true);
			out.writeInt(inputChanges.getDurationPort().getId());
		} else {
			out.writeBoolean(false);
			out.writeBoolean(inputChanges.isHiddenTimerTrigger());
		}
	}
	
	private void writeOutputEvolution(DataOutputStream out, OutputEvolution outputEvolution) throws IOException {
		out.writeInt(outputEvolution.getEvolution().size());
		
		for (Map<Port, String> m : outputEvolution.getEvolution()) {
			out.writeInt(m.size());
			
			for (Map.Entry<Port, String> e2 : m.entrySet()) {
				out.writeInt(e2.getKey().getId());
				out.writeUTF(e2.getValue());
			}
		}
	}
	
	private void writeTransition(DataOutputStream out, Transition transition) throws IOException {
		out.writeInt(transition.getInputChanges1().getId());
		out.writeInt(transition.getProtoSet().getId());
		out.writeInt(transition.getOutputEvolutions().size());
		
		for (OutputEvolution evo : transition.getOutputEvolutions()) {
			out.writeInt(evo.getId());
		}
		
		out.writeInt(transition.getDecaFourStateCfgs().size());
		
		for (int cfg : transition.getDecaFourStateCfgs()) {
			out.writeInt(cfg);
		}
	}
	
	private void writeProto(DataOutputStream out, Proto proto) throws IOException {
		out.writeUTF(proto.getDescription());
		out.writeUTF(proto.getTinyLink());
	}
	
	private void writeProtoSet(DataOutputStream out, ProtoSet protoSet) throws IOException {
		out.writeInt(protoSet.getElems().size());
		
		for (Proto p : protoSet.getElems()) {
			out.writeInt(p.getId());
		}
	}
	
//	private void writeResponse(DataOutputStream out, Response response) throws IOException {
//		out.writeInt(response.getEvos().size());
//		
//		for (OutputEvolution evo : response.getEvos()) {
//			out.writeInt(evo.getId());
//		}
//		
//		out.writeInt(response.getStatePerScope().size());
//		
//		for (Map.Entry<Scope, String> e : response.getStatePerScope().entrySet()) {
//			out.writeInt(e.getKey().getId());
//			out.writeUTF(e.getValue());
//		}
//		
//		out.writeInt(response.getValuation().size());
//		
//		for (Map.Entry<Port, String> e : response.getValuation().entrySet()) {
//			out.writeInt(e.getKey().getId());
//			out.writeUTF(e.getValue());
//		}
//	}
	
	private void writeCharSet(DataOutputStream out, CharSet charSet) throws IOException {
		out.writeInt(charSet.getVertex().getId());
		out.writeInt(charSet.getTraces().size());
		
		for (Trace trace : charSet.getTraces()) {
			writeTrace(out, trace);
		}
	}
	
	private void writeTrace(DataOutputStream out, Trace trace) throws IOException {
		out.writeInt(trace.getTransitions().size());
		
		for (Transition transition : trace.getTransitions()) {
			out.writeInt(transition.getId());
		}
	}
	
	private void writeTest(DataOutputStream out, Trace test) throws IOException {
		writeTrace(out, test);
	}
	
	private void writeTestSuite(DataOutputStream out, TestSuite testSuite) throws IOException {
		out.writeUTF(testSuite.getDescription());
		out.writeInt(testSuite.getTests().size());
		
		for (Trace test : testSuite.getTests()) {
			writeTest(out, test);
		}
	}
	
//	public static void main(String[] args) {
//		DecaStableFileWriter x = new DecaStableFileWriter();
//		x.loadFromFile("models", "all.reduced.2.stable", true);
//		x.saveToFile("models", "all.reduced.2.stable");
//	}
}

