package jlx.behave.stable.files;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jlx.asal.j.JInt;
import jlx.asal.j.JType;
import jlx.asal.vars.ASALPort;
import jlx.behave.stable.files.DecaStableFile.OutputEvolution;
import jlx.behave.stable.files.DecaStableFile.Port;
import jlx.behave.stable.files.DecaStableFile.Scope;
import jlx.behave.stable.files.DecaStableFile.Trace;
import jlx.behave.stable.files.DecaStableFile.Transition;
import jlx.blocks.ibd1.InPort;
import jlx.blocks.ibd1.OutPort;
import jlx.blocks.ibd1.PrimitivePort;
import jlx.common.reflection.ClassReflectionException;
import jlx.models.UnifyingBlock;
import jlx.scenario.Output;
import jlx.scenario.Scenario;
import jlx.scenario.Step;

public class DecaStableScenario extends Scenario {
	private final UnifyingBlock ub;
	private final StableToUnifyingBlock elements;
	private final Step[] steps;
	
	public DecaStableScenario(UnifyingBlock ub, Trace trace) throws DecaStableScenarioException {
		this.ub = ub;
		
		elements = StableToUnifyingBlock.getInstance(ub);
		steps = computeSteps(trace).toArray(new Step[0]);
	}
	
	/**
	 * This is messy, since we store values to suit the point software simulator... :-(
	 */
	private JType constructJType(Class<? extends JType> type, String v) throws DecaStableScenarioException {
		if (v.equals("\"999999999\"") && type.equals(JInt.class)) {
			return new JInt._600000();
		}
		
		String noQuotes = v.substring(1, v.length() - 1);
		
		try {
			return JType.createValue(ub.lib.getConstructor(v, type).getLegacy());
		} catch (ClassReflectionException e2) {
			try {
				String s = "\"" + type.getSimpleName() + "." + noQuotes + "\"";
				return JType.createValue(ub.lib.getConstructor(s, type).getLegacy());
			} catch (ClassReflectionException e) {
				throw new DecaStableScenarioException("Could not instantiate \"" + type.getSimpleName() + "." + noQuotes + "\" of type " + type.getCanonicalName() + "??", e);
			}
		}
	}
	
	private Output constructOutput(Map<Port, String> xs) throws DecaStableScenarioException {
		Map<OutPort<?>, JType> valuePerPort = new HashMap<OutPort<?>, JType>();
		
		for (Map.Entry<Port, String> x : xs.entrySet()) {
			ASALPort p = elements.getOutPort(x.getKey());
			JType v = constructJType(p.getType(), x.getValue());
			valuePerPort.put((OutPort<?>)p.getLegacy(), v);
		}
		
		return new Output(valuePerPort);
	}
	
	@SuppressWarnings("unchecked")
	private List<Step> computeSteps(Trace trace) throws DecaStableScenarioException {
		List<Step> result = new ArrayList<Step>();
		boolean first = true;
		
		for (Transition t : trace.getTransitions()) {
			if (t.getInputChanges1().isHiddenTimerTrigger()) {
				//Do nothing, since the model has no hidden (= unspecified) timeout triggers.
				//This is required to expose unspecified timeouts in the software simulator!!
				continue;
			}
			
			result.add(setMetadata(t));
			
			if (t.getInputChanges1().getDurationPort() != null) {
				PrimitivePort<?> p = elements.getDurationPort(t.getInputChanges1().getDurationPort()).getLegacy();
				result.add(triggerTimeout((InPort<?>)p));
			} else {
				for (Map.Entry<Port, String> e : t.getInputChanges1().getNewValuePerPort().entrySet()) {
					ASALPort p = elements.getInPort(e.getKey());
					
					if (elements.getDurationPorts().contains(p)) {
						//Duration ports are set to 99999999.
						//This is for the software simulator, do not do this for the specification FSM!
					} else {
						InPort<JType> ip = (InPort<JType>)p.getLegacy();
						JType v = constructJType(p.getType(), e.getValue());
						
						if (ip.getPossibleValues() != null) {
							if (!v.isElemIn(ip.getPossibleValues())) {
								throw new DecaStableScenarioException("Port \"" + e.getKey().getName() + "\" should not be assigned value \"" + e.getValue() + "\"??");
							}
						}
						
						result.add(setInput(ip, v));
					}
				}
			}
			
			for (OutputEvolution evo : t.getOutputEvolutions()) {
				if (first || evo.getEvolution().size() > 1) {
					Output initial = constructOutput(evo.getEvolution().get(0));
					Output[] evolution = new Output[evo.getEvolution().size() - 1];
					
					for (int i = 0; i < evolution.length; i++) {
						evolution[i] = constructOutput(evo.getEvolution().get(i + 1));
					}
					
					result.add(expectOutput(initial, evolution));
					first = false;
				}
			}
			
			if (t.getSrc() != null) {
				for (Map.Entry<Scope, String> e : t.getTgt().getStatePerScope().entrySet()) {
					String s1 = t.getSrc().getStatePerScope().get(e.getKey());
					String s2 = e.getValue();
					
					if (s1.equals(s2)) {
						result.add(expectStateChange(elements.getBlock(e.getKey()), s1));
					} else {
						result.add(expectStateChange(elements.getBlock(e.getKey()), s1, s2));
					}
				}
			} else {
				for (Map.Entry<Scope, String> e : t.getTgt().getStatePerScope().entrySet()) {
					result.add(expectStateChange(elements.getBlock(e.getKey()), e.getValue()));
				}
			}
			
			result.add(stabilize());
		}
		
		return result;
	}
	
	@Override
	public Step[] getSteps() {
		return steps;
	}
}

