package jlx.behave.stable.files;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import jlx.asal.j.JScope;
import jlx.asal.vars.ASALPort;
import jlx.behave.proto.DecaFourStateConfig;
import jlx.behave.proto.DecaFourVertex;
import jlx.behave.stable.files.DecaStableFile.Port;
import jlx.behave.stable.files.DecaStableFile.Scope;
import jlx.behave.stable.files.DecaStableFile.Vertex;
import jlx.blocks.ibd1.InPort;
import jlx.blocks.ibd1.OutPort;
import jlx.blocks.ibd1.Type1IBD;
import jlx.models.UnifyingBlock;
import jlx.models.UnifyingBlock.ReprBlock;
import jlx.models.UnifyingBlock.ReprPort;
import jlx.utils.Texts;

public class StableToUnifyingBlock {
	private final UnifyingBlock ub;
	private final Set<ASALPort> durationPorts;
	private final Map<Scope, Type1IBD> blockPerScope;
	private final Map<Port, ASALPort> portPerPort;
	private final Map<Port, ASALPort> portPerInput;
	private final Map<Port, ASALPort> portPerDuration;
	private final Map<Port, ASALPort> portPerOutput;
	
	private static Map<UnifyingBlock, StableToUnifyingBlock> instancePerUb = new HashMap<>();
	
	public static StableToUnifyingBlock getInstance(UnifyingBlock ub) {
		StableToUnifyingBlock result = instancePerUb.get(ub);
		
		if (result == null) {
			result = new StableToUnifyingBlock(ub);
			instancePerUb.put(ub, result);
		}
		
		return result;
	}
	
	private StableToUnifyingBlock(UnifyingBlock ub) {
		this.ub = ub;
		this.durationPorts = Collections.unmodifiableSet(extractDurationPorts());
		this.blockPerScope = new HashMap<Scope, Type1IBD>();
		this.portPerPort = new HashMap<Port, ASALPort>();
		this.portPerInput = new HashMap<Port, ASALPort>();
		this.portPerDuration = new HashMap<Port, ASALPort>();
		this.portPerOutput = new HashMap<Port, ASALPort>();
	}
	
	private Set<ASALPort> extractDurationPorts() {
		Set<ASALPort> result = new HashSet<ASALPort>();
		
		for (Map.Entry<ASALPort, ReprPort> e : ub.reprPortPerNarrowPort.entrySet()) {
			if (ub.sms4.timeoutPortPerDurationPort.containsKey(e.getValue())) {
				result.add(e.getKey());
			}
		}
		
		return result;
	}
	
	public Set<ASALPort> getDurationPorts() {
		return durationPorts;
	}
	
	public Type1IBD getBlock(Scope scope) {
		Type1IBD result = blockPerScope.get(scope);
		
		if (result != null) {
			return result;
		}
		
		List<String> names = new ArrayList<String>();
		
		for (Map.Entry<Type1IBD, ReprBlock> e : ub.reprBlockPerNarrowBlock.entrySet()) {
			if (e.getValue().getName().equals(scope.getName())) {
				blockPerScope.put(scope, e.getKey());
				return e.getKey();
			}
			
			names.add(e.getValue().getName());
		}
		
		throw new Error("Could not resolve scope named \"" + scope.getName() + "\" (available: " + Texts.concat(names, ", ") + ")!");
	}
	
	public ASALPort getPort(Port port) {
		ASALPort result = portPerPort.get(port);
		
		if (result != null) {
			return result;
		}
		
		Type1IBD block = getBlock(port.getOwner());
		result = block.getPrimitivePortPerName().get(port.getName());
		portPerPort.put(port, result);
		
		if (result == null) {
			throw new Error("Could not resolve port named \"" + port.getName() + "\"!");
		}
		
		return result;
	}
	
	private ASALPort getPort(Port port, Class<?> portType) {
		ASALPort result = getPort(port);
		
		if (result.getLegacy() == null) {
			throw new Error("Port \"" + port.getName() + "\" resolves to a helper port?!");
		}
		
		if (!portType.isAssignableFrom(result.getLegacy().getClass())) {
			throw new Error("Port \"" + port.getName() + "\" should resolve to " + portType.getCanonicalName() + ", but resolves to a " + result.getLegacy().getClass().getCanonicalName() + "?!");
		}
		
		return result;
	}
	
	public ASALPort getInPort(Port port) throws DecaStableScenarioException {
		ASALPort result = portPerInput.get(port);
		
		if (result == null) {
			result = getPort(port, InPort.class);
			
			if (!result.getType().getCanonicalName().equals(port.getTypeName())) {
				throw new Error("Port \"" + port.getName() + "\" should be of type " + result.getType().getCanonicalName() + ", but is of type " + port.getTypeName() + "?!");
			}
			
			portPerInput.put(port, result);
		}
		
		return result;
	}
	
	public ASALPort getDurationPort(Port port) throws DecaStableScenarioException {
		ASALPort result = portPerDuration.get(port);
		
		if (result == null) {
			result = getPort(port, InPort.class);
			
			if (!durationPorts.contains(result)) {
				throw new Error("Port \"" + port.getName() + "\" should be a duration port, but is a port of type " + port.getTypeName() + "?!");
			}
			
			portPerDuration.put(port, result);
		}
		
		return result;
	}
	
	public ASALPort getOutPort(Port port) throws DecaStableScenarioException {
		ASALPort result = portPerOutput.get(port);
		
		if (result == null) {
			result = getPort(port, OutPort.class);
			portPerOutput.put(port, result);
		}
		
		return result;
	}
	
//	public DecaFourVertex getVertex(String name) {
//		
//	}
//	
//	public DecaFourStateConfig getCfg(Vertex vertex) {
//		Map<JScope, DecaFourVertex> vertexPerScope = new HashMap<JScope, DecaFourVertex>();
//		
//		for (Map.Entry<Scope, String> e : vertex.getStatePerScope().entrySet()) {
//			JScope scope = ub.reprBlockPerType1IBD.get(getBlock(e.getKey()));
//			vertexPerScope.put(scope, getVertex(e.getValue()));
//		}
//		
//		return new DecaFourStateConfig(vertexPerScope);
//	}
}
