package jlx.scenario;

import java.util.*;

import jlx.asal.j.JType;
import jlx.asal.vars.ASALPort;
import jlx.behave.stable.files.DecaStableFile.Transition;
import jlx.blocks.ibd1.*;
import jlx.common.FileLocation;
import jlx.models.UnifyingBlock;
import jlx.models.UnifyingBlock.ReprPort;
import jlx.utils.Texts;

public abstract class Step {
	private final FileLocation fileLocation;
	
	private Step() {
		fileLocation = new FileLocation();
	}
	
	public FileLocation getFileLocation() {
		return fileLocation;
	}
	
	public abstract String oneLine(UnifyingBlock ub);
	
	public static class InputStep extends Step {
		private InPort<?> port;
		private JType newValue;
		
		public <T extends JType, V extends T> InputStep(InPort<T> port, V newValue) {
			this.port = port;
			this.newValue = newValue;
		}
		
		public InPort<?> getPort() {
			return port;
		}
		
		public JType getNewValue() {
			return newValue;
		}
		
		@Override
		public String oneLine(UnifyingBlock ub) {
			ReprPort rp = ub.reprPortPerPrimPort.get(port);
			
			if (rp.isDurationPort()) {
				return "Set timeout port " + rp.getName() + " to " + newValue.toStr();
			}
			
			return "Set input port " + rp.getName() + " to " + newValue.toStr();
		}
	}
	
	public static class TimeoutStep extends Step {
		private InPort<?> durationPort;
		
		public <T extends JType, V extends T> TimeoutStep(InPort<T> durationPort) {
			this.durationPort = durationPort;
		}
		
		public InPort<?> getDurationPort() {
			return durationPort;
		}
		
		@Override
		public String oneLine(UnifyingBlock ub) {
			ReprPort rp = ub.reprPortPerPrimPort.get(durationPort);
			ASALPort timeoutPort = ub.sms4.timeoutPortPerDurationPort.get(rp);
			return "Triggering timeout " + timeoutPort.getName();
		}
	}
	
	public static class OutputStep extends Step {
		private List<Map<OutPort<?>, JType>> outputSeq;
		
		public OutputStep(List<Map<OutPort<?>, JType>> outputSeq) {
			this.outputSeq = Collections.unmodifiableList(outputSeq);
		}
		
		public List<Map<OutPort<?>, JType>> getOutputSeq() {
			return outputSeq;
		}
		
		@Override
		public String oneLine(UnifyingBlock ub) {
			List<String> xs = new ArrayList<String>();
			
			for (Map<OutPort<?>, JType> output : outputSeq) {
				List<String> ys = new ArrayList<String>();
				
				for (Map.Entry<OutPort<?>, JType> e : output.entrySet()) {
					ReprPort rp = ub.reprPortPerPrimPort.get(e.getKey());
					ys.add(rp.getName() + "=" + e.getValue().toStr());
				}
				
				xs.add("(" + Texts.concat(ys, ", ") + ")");
			}
			
			return "Expect output evolution: " + Texts.concat(xs, " -> ");
		}
	}
	
	public static class StabilizeStep extends Step {
		@Override
		public String oneLine(UnifyingBlock ub) {
			return "Stabilize";
		}
	}
	
	public static class SingleStep extends Step {
		@Override
		public String oneLine(UnifyingBlock ub) {
			return "Do a single step";
		}
	}
	
	public static class StateChange extends Step {
		private final Type1IBD blockInstance;
		private final List<String> stateSeq;
		
		public StateChange(Type1IBD blockInstance, String state0, String... states) {
			this.blockInstance = blockInstance;
			
			stateSeq = new ArrayList<String>();
			stateSeq.add(state0);
			
			for (String state : states) {
				stateSeq.add(state);
			}
		}
		
		public Type1IBD getBlockInstance() {
			return blockInstance;
		}
		
		public List<String> getStateSeq() {
			return stateSeq;
		}
		
		@Override
		public String oneLine(UnifyingBlock ub) {
			return "Expect state change: " + Texts.concat(stateSeq, " -> ");
		}
	}
	
	public static class MetadataStep extends Step {
		private final Transition stableTransition;
		
		public MetadataStep(Transition stableTransition) {
			this.stableTransition = stableTransition;
		}
		
		public Transition getStableTransition() {
			return stableTransition;
		}
		
		@Override
		public String oneLine(UnifyingBlock ub) {
			if (getStableTransition().getSrc() != null) {
				return "(metadata)(id::" + stableTransition.getId() + ") From stable vertex " + getStableTransition().getSrc().getId() + " to " + getStableTransition().getTgt().getId();
			}
			
			return "(metadata)(id::" + stableTransition.getId() + ") From stable vertex NULL to " + getStableTransition().getTgt().getId();
		}
	}
}

