package jlx.utils.ldd;

import java.util.*;
import java.util.function.Consumer;

import jlx.utils.ldd.ops.*;

public class LDD<K, V> {
	private final LDDCache<K, V> cache;
	private final LDDNode rootNode;
	private final int hashCode;
	
	/**
	 * Creates LDD with the given elements.
	 * A new LDD encoding is created from the given possible values especially for the created LDD.
	 * @param elems
	 * @param pvsPerKey
	 */
	public LDD(Collection<Map<K, V>> elems, Map<K, ? extends Collection<V>> pvsPerKey) {
		this(elems, new LDDEncoding<K, V>(pvsPerKey));
	}
	
	/**
	 * Creates LDD without elements.
	 * @param elems
	 * @param encoding
	 */
	public LDD(LDDEncoding<K, V> encoding) {
		this(Collections.emptySet(), encoding);
	}
	
	/**
	 * Creates LDD with the given elements.
	 * @param elems
	 * @param encoding
	 */
	public LDD(Collection<Map<K, V>> elems, LDDEncoding<K, V> encoding) {
		cache = new LDDCache<K, V>(encoding);
		rootNode = cache.createNode(elems);
		cache.refresh(this);
		hashCode = extractHashCode();
	}
	
	/**
	 * Creates LDD that contains all combinations of the given possible values.
	 * @param encoding
	 * @param pvsPerKey
	 */
	public LDD(LDDEncoding<K, V> encoding, Map<K, ? extends Collection<V>> pvsPerKey) {
		cache = new LDDCache<K, V>(encoding);
		rootNode = isEmpty(pvsPerKey) ? null : LDDSetAll.setAll(cache, cache.createNode(Collections.emptyMap()), encoding.encode(pvsPerKey));
		cache.refresh(this);
		hashCode = extractHashCode();
	}
	
	private static <K, V> boolean isEmpty(Map<K, ? extends Collection<V>> pvsPerKey) {
		for (Map.Entry<K, ? extends Collection<V>> e : pvsPerKey.entrySet()) {
			if (e.getValue().isEmpty()) {
				return true;
			}
		}
		
		return pvsPerKey.isEmpty();
	}
	
	private LDD(LDDNode rootNode, LDDCache<K, V> cache) {
		this.rootNode = rootNode;
		this.cache = cache;
		this.cache.refresh(this);
		this.hashCode = extractHashCode();
	}
	
	private int extractHashCode() {
		return Objects.hash(rootNode, cache.getEncoding());
	}
	
	public int getNodeCount() {
		return cache.getNodeCount();
	}
	
	public LDDEncoding<K, V> getEncoding() {
		return cache.getEncoding();
	}
	
	public LDDNode getRootNode() {
		return rootNode;
	}
	
	public boolean isEmpty() {
		return rootNode == null;
	}
	
	public int size() {
		return rootNode != null ? LDDSize.getSize(rootNode) : 0;
	}
	
	public boolean isWellFormed() {
		return rootNode == null || rootNode.isWellFormed(cache.getEncoding());
	}
	
	public Map<K, V> getSomeElem() {
		if (rootNode == null) {
			return null;
		}
		
		return LDDFlatten.flatten(getEncoding(), rootNode);
	}
	
	public Set<Map<K, V>> getElems() {
		if (rootNode == null) {
			return Collections.emptySet();
		}
		
		Set<Map<K, V>> result = new HashSet<Map<K, V>>();
		LDDFlatten.flatten(getEncoding(), rootNode, new HashMap<K, V>(), result);
		return result;
	}
	
	public void processElems(Consumer<Map<K, V>> fct) {
		if (rootNode == null) {
			return;
		}
		
		LDDFlatten.flatten(getEncoding(), rootNode, new HashMap<K, V>(), fct);
	}
	
	public LDD<K, V> deepCopy() {
		LDDCache<K, V> newCache = new LDDCache<K, V>(getEncoding());
		return new LDD<K, V>(newCache.deepCopyNode(rootNode), newCache);
	}
	
	public LDD<K, V> setElems(Map<K, V> map) {
		LDDCache<K, V> newCache = new LDDCache<K, V>(getEncoding());
		return new LDD<K, V>(newCache.createNode(map), newCache);
	}
	
	public LDD<K, V> setElems(Collection<Map<K, V>> maps) {
		LDDCache<K, V> newCache = new LDDCache<K, V>(getEncoding());
		return new LDD<K, V>(newCache.createNode(maps), newCache);
	}
	
	public LDD<K, V> createSubLDD(Set<K> newKeys) {
		LDDEncoding<K, V> newEncoding = getEncoding().createSubEncoding(newKeys);
		LDDCache<K, V> newCache = new LDDCache<K, V>(newEncoding);
		LDDNode newRootNode;
		
		if (rootNode != null) {
			newRootNode = LDDProject.project(newCache, rootNode, getEncoding().encode(newKeys), 0);
		} else {
			newRootNode = null;
		}
		
		return new LDD<K, V>(newRootNode, newCache);
	}
	
	public LDD<K, V> clearAll(Set<K> vars) {
		LDDCache<K, V> newCache = new LDDCache<K, V>(getEncoding());
		LDDNode newRootNode;
		
		if (rootNode != null) {
			newRootNode = LDDClearAll.clearAll(newCache, rootNode, getEncoding().encode(vars));
		} else {
			newRootNode = newCache.createNode(Collections.emptyMap());
		}
		
		return new LDD<K, V>(newRootNode, newCache);
	}
	
	public LDD<K, V> setAll(Map<K, ? extends Collection<V>> newValues) {
		Map<Integer, List<Integer>> x = getEncoding().encode(newValues);
		LDDCache<K, V> newCache = new LDDCache<K, V>(getEncoding());
		LDDNode newRootNode;
		
		if (x.isEmpty()) {
			newRootNode = newCache.deepCopyNode(rootNode);
		} else {
			if (rootNode != null) {
				newRootNode = LDDSetAll.setAll(newCache, rootNode, x);
			} else {
				newRootNode = LDDSetAll.setAll(newCache, newCache.createNode(Collections.emptyMap()), x);
			}
		}
		
		return new LDD<K, V>(newRootNode, newCache);
	}
	
	public LDD<K, V> diff(Map<K, V> base, int diffCount) {
		return diff(base, diffCount, diffCount);
	}
	
	public LDD<K, V> diff(Map<K, V> base, int minDiffCount, int maxDiffCount) {
		LDDCache<K, V> newCache = new LDDCache<K, V>(getEncoding());
		LDDNode newRootNode;
		
		if (rootNode != null) {
			Map<Integer, Set<Integer>> excludedValuesPerKey = new HashMap<Integer, Set<Integer>>();
			
			for (Map.Entry<K, List<V>> e : getEncoding().getOrderedValuesPerKey().entrySet()) {
				Set<Integer> excludedValues = new HashSet<Integer>();
				V baseValue = base.get(e.getKey());
				
				if (baseValue != null) {
					excludedValues.add(-1);
					
					for (int i = 0; i < e.getValue().size(); i++) {
						if (!e.getValue().get(i).equals(baseValue)) {
							excludedValues.add(i);
						}
					}
				} else {
					for (int i = 0; i < e.getValue().size(); i++) {
						excludedValues.add(i);
					}
				}
				
				excludedValuesPerKey.put(getEncoding().encode(e.getKey()), excludedValues);
			}
			
			newRootNode = LDDDiffElems.diffElems(newCache, rootNode, excludedValuesPerKey, 0, minDiffCount, maxDiffCount);
		} else {
			newRootNode = rootNode;
		}
		
		return new LDD<K, V>(newRootNode, newCache);
	}
	
	public LDD<K, V> intersect(LDD<K, V> diagram) {
		return __intersect(diagram, Integer.MAX_VALUE);
	}
	
	public LDD<K, V> intersect(LDD<K, V> diagram, K maxIntersectionKey) {
		return __intersect(diagram, getEncoding().encode(maxIntersectionKey));
	}
	
	private LDD<K, V> __intersect(LDD<K, V> diagram, int maxIntersectionKeyIndex) {
		LDDCache<K, V> newCache = new LDDCache<K, V>(getEncoding());
		LDDNode newRootNode;
		
		if (rootNode != null && diagram.rootNode != null) {
			newRootNode = LDDIntersection.intersect(newCache, rootNode, diagram.rootNode, maxIntersectionKeyIndex);
		} else {
			newRootNode = null;
		}
		
		return new LDD<K, V>(newRootNode, newCache);
	}
	
	public LDD<K, V> union(LDD<K, V> diagram) {
		LDDCache<K, V> newCache = new LDDCache<K, V>(getEncoding());
		LDDNode newRootNode;
		
		if (rootNode != null) {
			if (diagram.rootNode != null) {
				newRootNode = LDDUnion.union(newCache, rootNode, diagram.rootNode);
			} else {
				newRootNode = rootNode;
			}
		} else {
			newRootNode = diagram.rootNode;
		}
		
		return new LDD<K, V>(newRootNode, newCache);
	}
	
	public boolean isDisjoint(LDD<K, V> diagram) {
		if (rootNode != null && diagram.rootNode != null) {
			return LDDDisjoint.isDisjoint(rootNode, diagram.rootNode);
		}
		
		return true;
	}
	
	public boolean contains(Map<K, V> elem) {
		return containsAll(new LDD<K, V>(Collections.singleton(elem), getEncoding()));
	}
	
	public boolean containsAll(LDD<K, V> diagram) {
		if (rootNode != null) {
			if (diagram.rootNode != null) {
				return LDDContainsAll.containsAll(rootNode, diagram.rootNode);
			} else {
				return true; //Other diagram is empty.
			}
		} else {
			//Other diagram must also be empty:
			return diagram.rootNode == null;
		}
	}
	
	public LDD<K, V> removeAll(LDD<K, V> diagram) {
		LDDCache<K, V> newCache = new LDDCache<K, V>(getEncoding());
		LDDNode newRootNode;
		
		if (rootNode != null) {
			if (diagram.rootNode != null) {
				newRootNode = LDDRemoveAll.removeAll(newCache, rootNode, diagram.rootNode);
			} else {
				newRootNode = rootNode; //Other diagram is empty.
			}
		} else {
			newRootNode = null;
		}
		
		return new LDD<K, V>(newRootNode, newCache);
	}
	
	@Override
	public int hashCode() {
		return hashCode;
	}
	
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		LDD<?, ?> other = (LDD<?, ?>) obj;
		return Objects.equals(getEncoding(), other.getEncoding()) && Objects.equals(rootNode, other.rootNode);
	}
	
	@Override
	public String toString() {
		return getEncoding().nodeToString(rootNode, true);
	}
	
	public LDDGraphVizFile toGraphVizFile() {
		return cache.toGraphVizFile(rootNode);
	}
}

