package jlx.utils.ldd;

import java.util.*;

import jlx.utils.ldd.ops.*;

public class LDDCache<K, V> {
	private final LDDEncoding<K, V> encoding;
	private final Map<LDDNode, LDDNode> nodeCache;
	
	public LDDCache(LDDEncoding<K, V> encoding) {
		this.encoding = encoding;
		
		nodeCache = new HashMap<LDDNode, LDDNode>();
	}
	
	public int getNodeCount() {
		return nodeCache.size();
	}
	
	public void refresh(LDD<K, V> diagram) {
		nodeCache.clear();
		refresh(diagram.getRootNode());
	}
	
	private void refresh(LDDNode node) {
		if (node != null) {
			getOrAddNode(node);
			refresh(node.getNextValueNode());
			refresh(node.getNextKeyNode());
		}
	}
	
	public LDDEncoding<K, V> getEncoding() {
		return encoding;
	}
	
	public Collection<LDDNode> getNodes() {
		return nodeCache.values();
	}
	
	public Set<List<LDDNode>> computeNodeSeqs() {
		Map<LDDNode, Set<LDDNode>> leftNodesPerNode = new HashMap<LDDNode, Set<LDDNode>>();
		
		for (LDDNode node : nodeCache.keySet()) {
			if (node.getNextValueNode() != null) {
				Set<LDDNode> leftNodes = leftNodesPerNode.get(node.getNextValueNode());
				
				if (leftNodes == null) {
					leftNodes = new HashSet<LDDNode>();
					leftNodesPerNode.put(node.getNextValueNode(), leftNodes);
				}
				
				leftNodes.add(node);
			}
		}
		
		Map<LDDNode, List<LDDNode>> nodeSeqPerNode = new HashMap<LDDNode, List<LDDNode>>();
		
		for (LDDNode node : nodeCache.keySet()) {
			Set<LDDNode> leftNodes = leftNodesPerNode.get(node);
			
			if (leftNodes == null || leftNodes.size() > 1) {
				List<LDDNode> nodeSeq = new ArrayList<LDDNode>();
				nodeSeq.add(node);
				
				LDDNode curr = node.getNextValueNode();
				
				while (curr != null && leftNodesPerNode.get(curr).size() == 1) {
					nodeSeq.add(curr);
					curr = curr.getNextValueNode();
				}
				
				nodeSeqPerNode.put(node, nodeSeq);
			}
		}
		
		return new HashSet<List<LDDNode>>(nodeSeqPerNode.values());
	}
	
	public LDDNode createNode(Collection<Map<K, V>> elems) {
		if (elems.isEmpty()) {
			return null;
		}
		
		Iterator<Map<K, V>> q = elems.iterator();
		LDDNode result = createNode(q.next());
		
		while (q.hasNext()) {
			result = LDDUnion.union(this, result, createNode(q.next()));
		}
		
		return result;
	}
	
	public LDDNode createNode(Map<K, V> elem) {
		return createSubSingleton(0, elem);
	}
	
	private LDDNode createSubSingleton(int topKeyId, Map<K, V> elem) {
		if (topKeyId < getEncoding().getOrderedKeys().size()) {
			K key = getEncoding().decode(topKeyId);
			int valueId = getEncoding().encode(key, elem.get(key));
			return getOrAddNode(new LDDNode(topKeyId, valueId, null, createSubSingleton(topKeyId + 1, elem)));
		}
		
		return null;
	}
	
	private LDDNode getOrAddNode(LDDNode node) {
		LDDNode result = nodeCache.get(node);
		
		if (result == null) {
			nodeCache.put(node, node);
			result = node;
		}
		
		return result;
	}
	
	public LDDNode deepCopyNode(LDDNode node) {
		if (node == null) {
			return null;
		}
		
		LDDNode nextValueNode = deepCopyNode(node.getNextValueNode());
		LDDNode nextKeyNode = deepCopyNode(node.getNextKeyNode());
		return getOrAddNode(new LDDNode(node.getKey(), node.getValue(), nextValueNode, nextKeyNode));
	}
	
	/**
	 * 
	 * @param key
	 * @param value
	 * @param nextValueNode Should already be in the cache!
	 * @param nextKeyNode Should already be in the cache!
	 * @return
	 */
	public LDDNode createNode(int key, int value, LDDNode nextValueNode, LDDNode nextKeyNode) {
		if (nextValueNode != null && !nodeCache.containsKey(nextValueNode)) {
			throw new Error("Should not happen!");
		}
		
		if (nextKeyNode != null && !nodeCache.containsKey(nextKeyNode)) {
			throw new Error("Should not happen!");
		}
		
		return getOrAddNode(new LDDNode(key, value, nextValueNode, nextKeyNode));
	}
	
	private static int down(int c) {
		return c / 2;
	}
	
	private static int up(int c) {
		return (c + 255) / 2;
	}
	
	private static String[] getRandomColors() {
		int r = (int)(Math.random() * 256);
		int g = (int)(Math.random() * 256);
		int b = (int)(Math.random() * 256);
		
		switch ((int)(Math.random() * 3)) {
			case 0:
				r = 0;
				break;
			case 1:
				g = 0;
				break;
			case 2:
				b = 0;
				break;
		}
		
		String c1 = String.format("#%02X%02X%02X", down(r), down(g), down(b));
		String c2 = String.format("#%02X%02X%02X", up(up(r)), up(up(g)), up(up(b)));
		return new String[] { c1, c2 };
	}
	
	private static String escapeHTML(String s) {
		String result = s;
		result = result.replace("\"", "&quot;");
		result = result.replace("&", "&amp;");
		result = result.replace("<", "&leq;");
		result = result.replace(">", "&geq;");
		return result;
	}
	
	public LDDGraphVizFile toGraphVizFile(LDDNode rootNode) {
		Set<List<LDDNode>> nodeSeqs = computeNodeSeqs();
		
		Map<List<LDDNode>, String> namePerNodeSeq = new HashMap<List<LDDNode>, String>();
		Map<LDDNode, String> namePerNode = new HashMap<LDDNode, String>();
		Map<LDDNode, List<LDDNode>> nodeSeqPerNode = new HashMap<LDDNode, List<LDDNode>>();
		Map<K, Set<List<LDDNode>>> nodeSeqsPerKey = new HashMap<K, Set<List<LDDNode>>>();
		
		for (K key : getEncoding().getOrderedKeys()) {
			nodeSeqsPerKey.put(key, new HashSet<List<LDDNode>>());
		}
		
		for (List<LDDNode> nodeSeq : nodeSeqs) {
			String name = "Q" + namePerNodeSeq.size();
			namePerNodeSeq.put(nodeSeq, name);
			
			for (int index = 0; index < nodeSeq.size(); index++) {
				namePerNode.put(nodeSeq.get(index), "N" + index);
				nodeSeqPerNode.put(nodeSeq.get(index), nodeSeq);
			}
			
			K key = getEncoding().decode(nodeSeq.get(0).getKey());
			nodeSeqsPerKey.get(key).add(nodeSeq);
		}
		
		LDDGraphVizFile result = new LDDGraphVizFile();
		result.add("digraph G {");
		
		if (rootNode != null) {
			result.add("\tI [label=\"start\", shape=Mdiamond];");
		}
		
		for (Map.Entry<K, Set<List<LDDNode>>> e : nodeSeqsPerKey.entrySet()) {
			String[] colors = getRandomColors();
			
			for (List<LDDNode> nodeSeq : e.getValue()) {
				K key = getEncoding().decode(nodeSeq.get(0).getKey());
				String label = "<TABLE BORDER=\"0\" CELLBORDER=\"1\" CELLSPACING=\"0\"><TR>";
				label += "<TD BGCOLOR=\"" + colors[0] + "\"><FONT COLOR=\"white\">" + escapeHTML(key.toString()) + "</FONT></TD>";
				
				for (LDDNode node : nodeSeq) {
					V value = getEncoding().decode(key, node.getValue());
					label += "<TD PORT=\"" + namePerNode.get(node) + "\" BGCOLOR=\"" + colors[1] + "\">" + (value != null ? escapeHTML(value.toString()) : "&#x2205;") + "</TD>";
				}
				
				label += "</TR></TABLE>";
				result.add("\t" + namePerNodeSeq.get(nodeSeq) + " [label=<" + label + ">, shape=plaintext];");
			}
		}
		
		if (rootNode != null) {
			result.add("\tI -> " + namePerNodeSeq.get(nodeSeqPerNode.get(rootNode)) + ":" + namePerNode.get(rootNode) + " [style=solid];");
		}
		
		for (LDDNode node : namePerNode.keySet()) {
			if (node.getNextKeyNode() != null) {
				String src = namePerNodeSeq.get(nodeSeqPerNode.get(node)) + ":" + namePerNode.get(node);
				String tgt = namePerNodeSeq.get(nodeSeqPerNode.get(node.getNextKeyNode())) + ":" + namePerNode.get(node.getNextKeyNode());
				result.add("\t" + src + " -> " + tgt + " [style=solid];");
			}
			
			if (node.getNextValueNode() != null) {
				if (nodeSeqPerNode.get(node) != nodeSeqPerNode.get(node.getNextValueNode())) {
					String src = namePerNodeSeq.get(nodeSeqPerNode.get(node)) + ":" + namePerNode.get(node);
					String tgt = namePerNodeSeq.get(nodeSeqPerNode.get(node.getNextValueNode())) + ":" + namePerNode.get(node.getNextValueNode());
					result.add("\t" + src + " -> " + tgt + " [style=dashed];");
					result.addTag("dashed");
				}
			}
		}
		
		result.add("}");
		return result;
	}
}

