package jlx.utils.ldd;

import java.util.*;

public class LDDEncoding<K, V> {
	private final Map<K, Integer> idPerKey;
	private final List<K> orderedKeys;
	private final Map<K, Map<V, Integer>> idPerValuePerKey;
	private final Map<K, List<V>> orderedValuesPerKey;
	
	public LDDEncoding(Map<K, ? extends Collection<V>> pvsPerKey) {
		this(pvsPerKey, new ArrayList<K>(pvsPerKey.keySet()));
	}
	
	public LDDEncoding(Map<K, ? extends Collection<V>> pvsPerKey, Collection<K> orderedKeys) {
		idPerKey = new HashMap<K, Integer>();
		idPerValuePerKey = new HashMap<K, Map<V, Integer>>();
		
		Map<K, List<V>> __orderedValuesPerKey = new HashMap<K, List<V>>();
		List<K> __orderedKeys = new ArrayList<K>();
		
		Iterator<K> q = orderedKeys.iterator();
		int index = 0;
		
		while (q.hasNext()) {
			K key = q.next();
			
			if (idPerKey.containsKey(key)) {
				throw new Error("Should not happen!");
			}
			
			Collection<V> values = pvsPerKey.get(key);
			
			if (values == null) {
				throw new Error("Should not happen!");
			}
			
			Map<V, Integer> idPerValue = new HashMap<V, Integer>();
			List<V> orderedValues = new ArrayList<V>();
			int nextId = 0;
			
			for (V value : values) {
				idPerValue.put(value, nextId);
				orderedValues.add(value);
				nextId++;
			}
			
			idPerKey.put(key, index);
			idPerValuePerKey.put(key, idPerValue);
			
			__orderedKeys.add(key);
			__orderedValuesPerKey.put(key, Collections.unmodifiableList(orderedValues));
			
			index++;
		}
		
		this.orderedKeys = Collections.unmodifiableList(__orderedKeys);
		this.orderedValuesPerKey = Collections.unmodifiableMap(__orderedValuesPerKey);
	}
	
	public LDDEncoding<K, V> createSubEncoding(Collection<K> newKeys) {
		if (!orderedKeys.containsAll(newKeys)) {
			throw new Error("Should not happen!");
		}
		
		Map<K, List<V>> newPvsPerKey = new HashMap<K, List<V>>();
		List<K> newOrderedKeys = new ArrayList<K>();
		
		for (K key : orderedKeys) {
			if (newKeys.contains(key)) {
				if (newOrderedKeys.contains(key)) {
					throw new Error("Should not happen!");
				}
				
				newPvsPerKey.put(key, orderedValuesPerKey.get(key));
				newOrderedKeys.add(key);
			}
		}
		
		return new LDDEncoding<K, V>(newPvsPerKey, newOrderedKeys);
	}
	
	public List<K> getOrderedKeys() {
		return orderedKeys;
	}
	
	public Map<K, List<V>> getOrderedValuesPerKey() {
		return orderedValuesPerKey;
	}
	
	public int encode(K key) {
		return idPerKey.get(key);
	}
	
	public K decode(int id) {
		return orderedKeys.get(id);
	}
	
	public int encode(K key, V value) {
		if (value == null) {
			return -1;
		}
		
		Integer result = idPerValuePerKey.get(key).get(value);
		
		if (result == null) {
			throw new Error("Could not encode " + value + "!");
		}
		
		return result;
	}
	
	private List<V> getOrderedValues(K key) {
		return orderedValuesPerKey.get(key);
	}
	
	public V decode(K key, int valueId) {
		return valueId >= 0 ? getOrderedValues(key).get(valueId) : null;
	}
	
	public boolean isValidKeyId(int keyId) {
		return keyId >= 0 && keyId < orderedKeys.size();
	}
	
	public boolean isValidValueId(int keyId, int valueId) {
		if (valueId < -1) {
			return false;
		}
		
		if (valueId >= getOrderedValues(decode(keyId)).size()) {
			return false;
		}
		
		return true;
	}
	
	public void checkElem(Map<K, V> elem) {
		for (Map.Entry<K, V> e : elem.entrySet()) {
			Map<V, Integer> idPerValue = idPerValuePerKey.get(e.getKey());
			
			if (idPerValue == null) {
				throw new Error("Invalid key: " + e.getKey());
			}
			
			if (!idPerValue.containsKey(e.getValue())) {
				throw new Error("Invalid value: " + e.getValue());
			}
		}
	}
	
	public Set<Integer> encode(Set<K> vars) {
		Set<Integer> result = new HashSet<Integer>();
		
		for (K var : vars) {
			result.add(encode(var));
		}
		
		return result;
	}
	
	public Map<Integer, List<Integer>> encode(Map<K, ? extends Collection<V>> map) {
		Map<Integer, List<Integer>> result = new HashMap<Integer, List<Integer>>();
		
//		for (K key : orderedKeys) {
//			Collection<V> values = map.get(key);
//			
//			if (values != null) {
//				SortedSet<Integer> valueIds = new TreeSet<Integer>();
//				
//				for (V value : values) {
//					valueIds.add(encode(key, value));
//				}
//				
//				result.put(encode(key), new ArrayList<Integer>(valueIds));
//			}
//		}
		
		for (Map.Entry<K, ? extends Collection<V>> e : map.entrySet()) {
			if (e.getValue().size() > 0) {
				SortedSet<Integer> values = new TreeSet<Integer>();
				
				for (V value : e.getValue()) {
					values.add(encode(e.getKey(), value));
				}
				
				result.put(encode(e.getKey()), new ArrayList<Integer>(values));
			}
		}
		
		return result;
	}
	
	public void putDecoded(Map<K, V> dest, int keyId, int valueId) {
		K key = decode(keyId);
		V value = decode(key, valueId);
		
		if (value != null) {
			dest.put(key, value);
		}
	}
	
	public String nodeToString(LDDNode node, boolean recursive) {
		if (node == null) {
			return "-";
		}
		
		K key = decode(node.getKey());
		String result = key.toString();
		result += "=" + String.valueOf(decode(key, node.getValue())) + "(" + node.getValue() + ")";
		
		if (recursive) {
			result += "," + nodeToString(node.getNextValueNode(), true);
			result += "," + nodeToString(node.getNextKeyNode(), true);
			result = "[" + result + "]";
		}
		
		return result;
	}
}

