package jlx.utils.ldd;

import java.util.*;

public class TransitiveLDD<K, V> {
	private LDD<WrappedKey<K>, V> ldd;
	private final List<WrappedKey<K>> srcKeys;
	private final List<WrappedKey<K>> tgtKeys;
	
	private static class WrappedKey<K> {
		public final K key;
		public final boolean isSrc;
		public final int hashCode;
		
		public WrappedKey(K key, boolean isSrc) {
			this.key = key;
			this.isSrc = isSrc;
			
			hashCode = Objects.hash(key, isSrc);
		}
		
		@Override
		public int hashCode() {
			return hashCode;
		}
		
		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			WrappedKey<?> other = (WrappedKey<?>) obj;
			return isSrc == other.isSrc && Objects.equals(key, other.key);
		}
	}
	
	private TransitiveLDD(TransitiveLDD<K, V> src, LDD<WrappedKey<K>, V> ldd) {
		this.ldd = ldd;
		
		srcKeys = src.srcKeys;
		tgtKeys = src.tgtKeys;
	}
	
	public TransitiveLDD(Map<K, ? extends Collection<V>> pvsPerKey, Collection<K> orderedKeys) {
		Map<WrappedKey<K>, List<V>> newPvsPerKey = new HashMap<WrappedKey<K>, List<V>>();
		srcKeys = new ArrayList<WrappedKey<K>>();
		tgtKeys = new ArrayList<WrappedKey<K>>();
		
		for (K key : orderedKeys) {
			Collection<V> pvs = pvsPerKey.get(key);
			
			if (pvs == null) {
				throw new Error("Should not happen!");
			}
			
			List<V> newPvs = new ArrayList<V>();
			
			for (V pv : pvs) {
				if (newPvs.contains(pv)) {
					throw new Error("Should not happen!");
				}
				
				newPvs.add(pv);
			}
			
			WrappedKey<K> newSrcKey = new WrappedKey<K>(key, true);
			WrappedKey<K> newTgtKey = new WrappedKey<K>(key, false);
			
			if (newPvsPerKey.containsKey(newSrcKey) || newPvsPerKey.containsKey(newTgtKey)) {
				throw new Error("Should not happen!");
			}
			
			newPvsPerKey.put(newSrcKey, newPvs);
			newPvsPerKey.put(newTgtKey, newPvs);
			srcKeys.add(newSrcKey);
			tgtKeys.add(newTgtKey);
		}
		
		List<WrappedKey<K>> keys = new ArrayList<WrappedKey<K>>(srcKeys);
		keys.addAll(tgtKeys);
		
		ldd = new LDD<WrappedKey<K>, V>(new LDDEncoding<WrappedKey<K>, V>(newPvsPerKey, keys));
	}
	
	public TransitiveLDD<K, V> add(Map<K, V> src, Map<K, V> tgt) {
		Map<WrappedKey<K>, V> x = new HashMap<WrappedKey<K>, V>();
		
		for (Map.Entry<K, V> e : src.entrySet()) {
			x.put(new WrappedKey<K>(e.getKey(), true), e.getValue());
		}
		
		for (Map.Entry<K, V> e : tgt.entrySet()) {
			x.put(new WrappedKey<K>(e.getKey(), false), e.getValue());
		}
		
		ldd = ldd.union(new LDD<WrappedKey<K>, V>(Collections.singleton(x), ldd.getEncoding()));
		return this;
		//return new TransitiveLDD<K, V>(this, ldd.union(new LDD<WrappedKey<K>, V>(Collections.singleton(x), ldd.getEncoding())));
	}
	
	public TransitiveLDD<K, V> add(TransitiveLDD<K, V> x) {
		return new TransitiveLDD<K, V>(this, ldd.union(x.ldd));
	}
	
	public Set<Map<K, V>> getTgts(Map<K, V> src) {
		Map<WrappedKey<K>, V> x = new HashMap<WrappedKey<K>, V>();
		
		for (Map.Entry<K, V> e : src.entrySet()) {
			x.put(new WrappedKey<K>(e.getKey(), true), e.getValue());
		}
		
		WrappedKey<K> lastSrcKey = srcKeys.get(srcKeys.size() - 1);
		Set<Map<K, V>> result = new HashSet<Map<K, V>>();
		
		for (Map<WrappedKey<K>, V> z : ldd.intersect(new LDD<WrappedKey<K>, V>(Collections.singleton(x), ldd.getEncoding()), lastSrcKey).getElems()) {
			Map<K, V> map = new HashMap<K, V>();
			
			for (Map.Entry<WrappedKey<K>, V> e : z.entrySet()) {
				map.put(e.getKey().key, e.getValue());
			}
			
			result.add(map);
		}
		
		return result;
	}
	
	public int getSize() {
		return ldd.size();
	}
}

